import pandas as pd
import typing as typ
import pathlib as pth
import dataframe_image as dfi
import pandas.io.formats.style as pdst

import pacifico_devel.util.reportTemplate.src.core.Plot.Plot as bplt
from pacifico_devel.util.reportTemplate.src.util.Patches import ScreenshotPatch
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl
import pacifico_devel.util.reportTemplate.src.util.PacificoPlotColors as pc

import pacifico_devel.util.lexikon.src.util.Enumerations as lenm
import pacifico_devel.util.lexikon.src.core.Translator as trl


# TO DO: Implement formatter

class PlotTableColorIndex(bplt.Plot):
    def __init__(self,
                 df: pd.DataFrame,
                 languageInput: lenm.Language,
                 languageOutput: lenm.Language,
                 shape: typ.Tuple[float, float],
                 basePlotResolution: int,
                 style: typ.Optional[stl.Style] = None,
                 formatter: typ.Optional[typ.Dict[str, str]] = None):
        super(PlotTableColorIndex, self).__init__(df=df,
                                                  shape=shape,
                                                  languageInput=languageInput,
                                                  languageOutput=languageOutput,
                                                  basePlotResolution=basePlotResolution,
                                                  style=style)
        self._formatter = formatter

    def saveImage(self) -> None:
        """


        Args:
            dfTranslated:

        Returns:

        """
        chart = self._makeChart()
        ScreenshotPatch.patch()  # Monkey patch for chrome screenshots server
        print(f"SAVING TABLE TO: {self.pathImage} ...")
        dfi.export(obj=chart,
                   filename=str(self.pathImage))
        print(f"SAVING TABLE TO: {self.pathImage} ... DONE!")

    def _makeChart(self) -> pdst.Styler:
        # dfAux = self._translateDataFrame(df)
        thProps = [("background-color", pc.PacificoPlotColors.BLUE.getCodeString()),
                   ("color", "white"),
                   ('font-family', 'Klartext Mono'),
                   # ('text-align', 'justify'),
                   ('font-size', '14pt')]

        tdProps = [('font-family', 'Klartext Mono'),
                   ('font-size', '14pt')]

        dfst = self.df.style.set_table_styles([{"selector": "th", "props": thProps},
                                               {"selector": "td", "props": tdProps}])
        if self._formatter is None:
            dfst = dfst.format("{:.2f}")
        else:
            dfst = dfst.format(self._formatter)
        return dfst

    def df_to_excel(self, writer, tag):
        """
        saves DataFrame to excel
        Returns:

        """

        self.df.to_excel(writer, sheet_name=tag)



from pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokeh import PlotBokeh
import pacifico_devel.util.reportTemplate.src.util.PacificoPlotColors as pc
import pacifico_devel.util.lexikon.src.util.Enumerations as lenm
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl

import bokeh.plotting as bpl
from bokeh.models import ColumnDataSource
from bokeh.plotting import figure
from bokeh.transform import cumsum
from bokeh.models import LabelSet

import pandas as pd
import numpy as np
import typing as typ
from bokeh.plotting import output_file, show
from bokeh.io import export_png


class PlotBokehPieChart(PlotBokeh):
    def __init__(self,
                 df: pd.DataFrame,
                 shape: typ.Tuple[float, float],
                 languageInput: lenm.Language,
                 languageOutput: lenm.Language,
                 basePlotResolution: int,
                 style: typ.Optional[stl.Style] = None,
                 title: str = None):
        super(PlotBokehPieChart, self).__init__(df=df, shape=shape, languageInput=languageInput,
                                                languageOutput=languageOutput, basePlotResolution=basePlotResolution,
                                                style=style)
        self._title = title

    def _makeChart(self) -> bpl.Figure:
        assets = list(self.df['component'])
        self.df['angle'] = self.df['weight'] / self.df['weight'].sum() * 2 * np.pi
        self.df['color'] = [color.getCodeString() for i, color in zip(range(len(assets)), pc.PacificoPlotColors)]
        chart = figure(width=self._shape[0],
                       height=self._shape[1],
                       toolbar_location=None)
        chart.title = self._title
        chart.wedge(x=0, y=1,
                    radius=0.75,
                    start_angle=cumsum('angle', include_zero=True),
                    end_angle=cumsum('angle'),
                    line_color="white",
                    fill_color='color',
                    legend_field='component',
                    source=self.df)
        self.df["weight"] = round(self.df['weight'], 1).astype(str) + ' %'
        self.df["weight"] = self.df["weight"].str.pad(20, side="left")
        source = ColumnDataSource(self.df)
        labels = LabelSet(x=0, y=1, text='weight', angle=cumsum('angle', include_zero=True),
                          source=source, render_mode='canvas', text_color='#FFFFFF', text_font_size="15px")
        chart.add_layout(labels)
        self.style(chart)
        self.legendStyle(chart, 'vertical', 'right')
        return chart

    def legendStyle(self, plot, orientation: str = 'vertical', layout: str = 'right'):
        new_legend = plot.legend[0]
        plot.add_layout(new_legend, layout)  # 'below'
        plot.legend.orientation = orientation  # 'horizontal'
        plot.legend.location = "center"
        plot.legend.border_line_alpha = 0
        plot.legend.background_fill_color = None
        plot.xgrid.grid_line_color = None
        plot.legend.click_policy = "hide"
        plot.legend.label_text_font = "klartext mono"
        plot.legend.label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        return plot

    def style(self, plot):
        plot.title.text_color = pc.PacificoPlotColors.BLUE.getCodeString()
        plot.title.text_font_size = "20px"
        plot.axis.axis_label = None
        plot.axis.visible = False
        plot.grid.grid_line_color = None
        plot.outline_line_color = None
        plot.legend.border_line_alpha = 0
        plot.legend.label_text_font = "klartext mono"
        return plot


if __name__ == '__main__':
    from datetime import datetime
    from Applications.ApplicationReportChileUnemployment.src.core.Data import Data
    from pacifico_devel.util.lexikon.src.util.Enumerations import Language
    import chromedriver_autoinstaller

    chromedriver_autoinstaller.install()
    data = Data(datetime.now())
    data.load('KmK9FqgJ5h1pSNjh1vhha6ul30vXBOBiMVlmplXf', ['60703000-6', 'F049.CES.TAS.UCH3.65.T'])
    data.computeUnemployment()
    data = data.getData()
    months = {'ene': 1, 'feb': 2, 'mar': 3, 'abr': 4, 'may': 5, 'jun': 6, 'jul': 7,
              'ago': 8, 'sep': 9, 'oct': 10, 'nov': 11, 'dic': 12}
    dataDict = {}
    years = list(set(data['Año']))
    years.sort()
    for year in years[-5:]:
        dataDict[str(year)] = data[data['Año'] == year]['Desempleo (Total)'].tolist()
    dataPlot = {**{'xAxisCategory': list(months.keys())}, **dataDict}

    dataPlot = pd.DataFrame(dataPlot)

    PlotBokehMultiBar(df=dataPlot,
                      shape=(900, 500),
                      languageInput=Language.fromDeeplString("ES"),
                      languageOutput=Language.fromDeeplString("ES"),
                      basePlotResolution=1000,
                      style=None,
                      title='')._makeChart()  # debugMakeChart(dataPlot, (900, 500))



"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

from pacifico.core.Service.ServiceRequest import ServiceRequest

class ServiceRequestApplication(ServiceRequest.ServiceRequest):
    def __init__(self, token, applicationName, applicationArguments, help):
        ServiceRequest.ServiceRequest.__init__(self)
        self.token = token
        self.applicationName = applicationName
        self.__setArguments(help, applicationArguments)

    # token
    def _get_token(self):
        return self.__token
    def _set_token(self, value):
        if not isinstance(value, str):
            raise TypeError("The token must be set to a string.")
        self.__token = value
    token = property(_get_token, _set_token)

    # applicationName
    def _get_applicationName(self):
        return self.__applicationName
    def _set_applicationName(self, value):
        if not isinstance(value, str):
            raise TypeError("The applicationName must be set to a string.")
        self.__applicationName = value
    applicationName = property(_get_applicationName, _set_applicationName)

    # applicationArguments
    def _get_applicationArguments(self):
        return self.__applicationArguments
    def _set_applicationArguments(self, value):
        if not isinstance(value, dict):
            raise TypeError("The applicationArguments must be set to a dictionary.")
        # if not all(isinstance(item, FieldQuality.FieldQuality) for item in value):
        #     raise TypeError("The quality must be set to a list of items of class 'FieldQuality.FieldQuality'.")
        self.__applicationArguments = value
    applicationArguments = property(_get_applicationArguments, _set_applicationArguments)

    # help
    def _get_help(self):
        return self.__help
    def _set_help(self, value):
        if not isinstance(value, bool):
            raise TypeError("The help must be set to a boolean.")
        self.__help = value
    help = property(_get_help, _set_help)

    def __setArguments(self, help, applicationArguments):
        self.help = help
        if self.help:
            self.applicationArguments = {}
        else:
            # Upload local argument files
            from pacifico.util import Format, Enumerations, Downloader
            from pacifico_devel.api.src.util import Enumerations as develEnumerations
            from pacifico_devel import Message
            auxServiceRequestApplication = ServiceRequestApplication(self.token, self.applicationName, applicationArguments, True)
            url = auxServiceRequestApplication.getResponse(self.token)
            timeOut = 300
            timeFrecuency = 0.5
            response = Downloader.downloadResponse(url, timeOut, timeFrecuency)
            format = 'dataFrame'
            format = Enumerations.Format.fromString(format)
            response = Format.jsonToFormat(format, response)
            response = response.loc[response['Subsection'] == 'Type']
            response['Value'] = response['Value'].apply(lambda x: develEnumerations.ValueType.fromString(x))
            for argumentName, valueType in zip(response['Section'], response['Value']):
                if valueType == develEnumerations.ValueType.ValueType_File:
                    # Upload the local file indicated by a file argument to a download url
                    localFilePath = applicationArguments.get(argumentName)
                    if localFilePath is not None:
                        url = Message.Report.uploadTemporalFileAndGetUrl(localFilePath)
                        applicationArguments[argumentName] = url
            # Set the arguments
            self.applicationArguments = applicationArguments

    def getToken(self):
        return self.token

    def getApplicationName(self):
        return self.applicationName

    def getArguments(self):
        return self.applicationArguments

    @staticmethod
    def create(applicationName, applicationArguments, help, token):
        return ServiceRequestApplication(token, applicationName, applicationArguments, help)


"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

from pacifico.core.Service.Field import Field
from pacifico.util import Enumerations
import datetime

class FieldTimeInterval(Field.Field):
    def __init__(self, dateStart=datetime.date.today(), dateEnd=datetime.date.today(), fixing=Enumerations.Fixing.EOD):
        if not dateEnd >= dateStart:
            raise ValueError('The dateEnd must be equal or superior than dateStart.')
        self.dateStart = dateStart
        self.dateEnd = dateEnd
        self.fixing = fixing

    # dateStart
    def _get_dateStart(self):
        return self.__dateStart
    def _set_dateStart(self, value):
        if not isinstance(value, datetime.date):
            raise TypeError("The dateStart must be set to a class 'datetime.date'.")
        self.__dateStart = value
    dateStart = property(_get_dateStart, _set_dateStart)

    # dateEnd
    def _get_dateEnd(self):
        return self.__dateEnd
    def _set_dateEnd(self, value):
        if not isinstance(value, datetime.date):
            raise TypeError("The dateEnd must be set to a class 'datetime.date'.")
        self.__dateEnd = value
    dateEnd = property(_get_dateEnd, _set_dateEnd)

    # fixing
    def _get_fixing(self):
        return self.__fixing
    def _set_fixing(self, value):
        if not isinstance(value, Enumerations.Fixing):
            raise TypeError("The fixing must be set to a class 'Enumerations.Fixing'.")
        self.__fixing = value
    fixing = property(_get_fixing, _set_fixing)

    # Class Methods

    def getDateStart(self):
        return self.dateStart

    def getDateEnd(self):
        return self.dateEnd

    def getFixing(self):
        return self.fixing


"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import json
from .. import ServiceRequest
from ...Field import FieldTicker, Field, FieldQuality, FieldTimeInterval, FieldDataType, FieldVersion, FieldReport


class ServiceRequestDataPoint(ServiceRequest.ServiceRequest):
    def __init__(self, instruments=[], timeHorizon=[], dataTypes=[], version=[], quality=[], optional={}):
        self.instruments = instruments # Tickers
        self.timeHorizon = timeHorizon
        self.dataTypes = dataTypes # Calculation Types
        self.version = version # Scenario
        self.quality = quality
        self.optional = optional

    # instruments
    def _get_instruments(self):
        return self.__instruments
    def _set_instruments(self, value):
        if not isinstance(value, list):
            raise TypeError("The instruments must be set to a list.")
        if not all(isinstance(item, (FieldTicker.FieldTicker, FieldReport.FieldReport)) for item in value):
            raise TypeError(
                "The instruments must be set to a list of items of class 'FieldTicker.FieldTicker' or 'FieldReport.FieldReport'.")
        self.__instruments = value
    instruments = property(_get_instruments, _set_instruments)

    # timeHorizon
    def _get_timeHorizon(self):
        return self.__timeHorizon
    def _set_timeHorizon(self, value):
        if not isinstance(value, list):
            raise TypeError("The timeHorizon must be set to a list.")
        if not all(isinstance(item, FieldTimeInterval.FieldTimeInterval) for item in value):
            raise TypeError("The timeHorizon must be set to a list of items of class "
                            "'FieldTimeInterval.FieldTimeInterval'.")
        self.__timeHorizon = value
    timeHorizon = property(_get_timeHorizon, _set_timeHorizon)

    # dataTypes
    def _get_dataTypes(self):
        return self.__dataTypes
    def _set_dataTypes(self, value):
        if not isinstance(value, list):
            raise TypeError("The dataTypes must be set to a list.")
        if not all(isinstance(item, FieldDataType.FieldDataType) for item in value):
            raise TypeError(
                "The dataTypes must be set to a list of items of class 'FieldDataType.FieldDataType'.")
        self.__dataTypes = value
    dataTypes = property(_get_dataTypes, _set_dataTypes)

    # version
    def _get_version(self):
        return self.__version
    def _set_version(self, value):
        if not isinstance(value, list):
            raise TypeError("The version must be set to a list.")
        if not all(isinstance(item, FieldVersion.FieldVersion) for item in value):
            raise TypeError(
                "The version must be set to a list of items of class 'FieldVersion.FieldVersion'.")
        self.__version = value
    version = property(_get_version, _set_version)

    # quality
    def _get_quality(self):
        return self.__quality
    def _set_quality(self, value):
        if not isinstance(value, list):
            raise TypeError("The quality must be set to a list.")
        if not all(isinstance(item, FieldQuality.FieldQuality) for item in value):
            raise TypeError(
                "The quality must be set to a list of items of class 'FieldQuality.FieldQuality'.")
        self.__quality = value
    quality = property(_get_quality, _set_quality)

    # optional
    def _get_optional(self):
        return self.__optional
    def _set_optional(self, value):
        if not isinstance(value, dict):
            raise TypeError("The optional must be set to a dictionary.")
        if not all(isinstance(item, Field.Field) for item in value.values()):
            raise TypeError("The optional must be set to a dictionary of items of class 'Field.Field'.")
        self.__optional = value
    optional = property(_get_optional, _set_optional)

    # Class Methods

    def getInstruments(self):
        return self.instruments

    def getReports(self):
        return self.instruments

    def getTimeHorizon(self):
        return self.timeHorizon

    def getDataTypes(self):
        return self.dataTypes

    def getVersions(self):
        return self.version

    def getQualities(self):
        return self.quality

    def getOptionals(self):
        return self.optional

    def getBase(self):
        for base in self.__class__.__bases__:
            return base.__name__

    @staticmethod
    def __serializeAndClean(dictionary):
        delete = []
        for key, value in zip(dictionary.keys(), dictionary.values()):
            if value == [] or value == {} or value == None:
                delete.append(key)
            elif isinstance(value, list):
                dictionary[key] = [item.getJson() for item in value]
            else:
                dictionary[key] = value
        for key in delete:
            del dictionary[key]
        return dictionary

    def getJson(self):
        # Get all attributes as a dictionary
        data = self.__dict__.copy()
        # Delete object private prefix from attribute names
        oldKeys = list(data.keys())
        for key in oldKeys:
            newKey = key.replace('_' + self.__class__.__name__ + '__', '')
            newKey = newKey.replace('_' + self.getBase() + '__', '')
            data[newKey] = data.pop(key)
        # Serialize object within the attribute and delete empty attributes
        data = self.__serializeAndClean(data)
        # jsonize data
        dataJson = json.dumps(data)
        return dataJson



"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import json
import pandas
import requests
import datetime
from . import Enumerations
from ....util.selenium import SeleniumWebdriver
from ..io.DataPoint.DataPointReport import DataPointReport as ioDataPointReport
from .... import Message


def jsonToFormat(format, response):
    if format == Enumerations.Format.Json:
        return response
    elif format == Enumerations.Format.Dictionary:
        return json.loads(response)
    elif format == Enumerations.Format.DataFrame:
        return jsonToDataFrame(response)


def jsonToDataFrame(jsonString):
    # Is a json value
    isJsonValue = '"field"' in jsonString
    # Is a json Report
    isJsonReport = '"variant"' in jsonString or '/bfi/input/reports' in jsonString
    if isJsonValue:
        return jsonValuesToDataFrame(jsonString)
    elif isJsonReport:
        return jsonReportsToDataFrame(jsonString)
    else:
        return jsonValuesToDataFrame(jsonString)


def jsonValuesToDataFrame(jsonString):
    dataFrameData = []
    data = json.loads(jsonString)
    for scenario in data.keys():
        for datePublication in data[scenario].keys():
            for dateEffective in data[scenario][datePublication].keys():
                for country in data[scenario][datePublication][dateEffective].keys():
                    for market in data[scenario][datePublication][dateEffective][country].keys():
                        for group in data[scenario][datePublication][dateEffective][country][market].keys():
                            for family in data[scenario][datePublication][dateEffective][country][market][group].keys():
                                for ticker in data[scenario][datePublication][dateEffective][country][market][group][
                                    family].keys():
                                    for dictValue in \
                                    data[scenario][datePublication][dateEffective][country][market][group][family][
                                        ticker]:
                                        value = dictValue['value']
                                        field = dictValue['field']
                                        if 'dateTenor' in dictValue.keys():
                                            if isinstance(dictValue['dateTenor'], str):
                                                dateTenor = datetime.datetime.strptime(dictValue['dateTenor'],
                                                                                       "%d/%m/%Y %H:%M:%S")
                                            else:
                                                dateTenor = dictValue['dateTenor']
                                        else:
                                            dateTenor = ''
                                        if 'other' in dictValue.keys():
                                            other = dictValue['other']
                                        else:
                                            other = ''
                                        if isinstance(dateEffective, str):
                                            dateEffectiveAux = datetime.datetime.strptime(dateEffective,
                                                                                          "%d/%m/%Y %H:%M:%S")
                                        else:
                                            dateEffectiveAux = dateEffective
                                        if isinstance(datePublication, str):
                                            datePublicationAux = datetime.datetime.strptime(datePublication, "%d/%m/%Y")
                                        else:
                                            datePublicationAux = datePublication
                                        dataFrameData.append(
                                            [scenario, datePublicationAux, dateEffectiveAux, country, market, group,
                                             family, ticker, value, field, dateTenor, other])
    return pandas.DataFrame(dataFrameData,
                            columns=['Scenario', 'Date Publication', 'Date Effective', 'Country', 'Market', 'Group',
                                     'Family', 'Ticker', 'Value', 'Field', 'Date Tenor', 'Other'])


def jsonReportsToDataFrame(jsonString):
    dataFrameData = []
    firstBrowser = True
    windowCount = 0
    if '/bfi/input/reports' in jsonString:
        reports = parseRawReports(jsonString)
        awsAccessKey = None
        awsSecretKey = None
        jsonString = ioDataPointReport.DataPointReportFromMessageReport(reports, awsAccessKey, awsSecretKey)
    data = json.loads(jsonString)
    if isinstance(data, list):  # Response is too big, over 5000 reports
        # print('Large response received, parsing...', end='')
        author = data[0]
        for reportJson in data[1:]:
            document = reportJson.get('document', '')
            chapter = reportJson.get('chapter', '')
            section = reportJson.get('section', '')
            subsection = reportJson.get('subsection', '')
            paragraph = reportJson.get('paragraph', '')
            item = reportJson.get('item', '')
            datePublicationObject = datetime.datetime.fromisoformat(reportJson.get('datePublication'))
            dateEffectiveObject = datetime.datetime.fromisoformat(reportJson.get('dateEffective'))
            fixing = reportJson.get('fixingPublication', '')
            variant = json.loads(reportJson.get('variant'))
            variantValueType = Enumerations.ValueType(int(variant['valueType']))
            variantValue = variantValueType.forceType(variant['value'])
            variantValueType = variantValueType.getString()
            dateTenor = reportJson.get('dateTenor')
            if dateTenor is None:
                dateTenor = ''
            else:
                dateTenor = datetime.datetime.fromisoformat(dateTenor)
            other = reportJson.get('comment', '')
            dataFrameData.append(
                [author, document, chapter, section, subsection, paragraph, item, datePublicationObject,
                 dateEffectiveObject, fixing, variantValue, variantValueType, dateTenor, other])
        # print('Done!')
    else:
        for author in data.keys():
            for datePublication in data[author].keys():
                for dateEffective in data[author][datePublication].keys():
                    for document in data[author][datePublication][dateEffective].keys():
                        for chapter in data[author][datePublication][dateEffective][document].keys():
                            for section in data[author][datePublication][dateEffective][document][chapter].keys():
                                for subsection in data[author][datePublication][dateEffective][document][chapter][
                                    section].keys():
                                    for paragraph in \
                                    data[author][datePublication][dateEffective][document][chapter][section][
                                        subsection].keys():
                                        for item in \
                                        data[author][datePublication][dateEffective][document][chapter][section][
                                            subsection][paragraph].keys():
                                            for reportDict in \
                                            data[author][datePublication][dateEffective][document][chapter][section][
                                                subsection][paragraph][item]:
                                                if isinstance(datePublication, str):
                                                    datePublicationObject = datetime.datetime.strptime(datePublication,
                                                                                                       "%d/%m/%Y")
                                                else:
                                                    datePublicationObject = datePublication
                                                if 'fixing' in reportDict.keys():
                                                    fixing = reportDict['fixing']
                                                else:
                                                    fixing = ''
                                                if isinstance(dateEffective, str):
                                                    dateEffectiveObject = datetime.datetime.strptime(dateEffective,
                                                                                                     "%d/%m/%Y %H:%M:%S")
                                                else:
                                                    dateEffectiveObject = dateEffective
                                                variant = reportDict['variant']
                                                variantValueType = Enumerations.ValueType.fromString(variant['type'])
                                                variantValue = variantValueType.forceType(variant['value'])
                                                if variantValueType == Enumerations.ValueType.ValueType_Browser:
                                                    try:
                                                        import os
                                                        from selenium.webdriver.common.keys import Keys
                                                        html = requests.get(variantValue).content
                                                        wd = os.getcwd()
                                                        filename = os.path.join(wd, 'browser.html')
                                                        with open(filename, 'wb') as file:
                                                            file.write(html)
                                                        if firstBrowser:
                                                            driver = SeleniumWebdriver.getDriver(detach=True)
                                                            firstBrowser = False
                                                        else:
                                                            windowCount += 1
                                                            tabName = str(windowCount)
                                                            driver.execute_script(
                                                                f"window.open('about:blank','{tabName}');")
                                                            driver.switch_to.window(f"{tabName}")
                                                        driver.get("file:" + filename)
                                                        os.remove(filename)
                                                    except Exception as e:
                                                        print(f"Formatting Error: Browser type values couldn't be "
                                                              f"displayed, probably because Google Chrome is not "
                                                              f"installed. ({str(e)})")
                                                variantValueType = variantValueType.getString()
                                                if 'dateTenor' in reportDict.keys():
                                                    if isinstance(reportDict['dateTenor'], str):
                                                        dateTenor = datetime.datetime.strptime(reportDict['dateTenor'],
                                                                                               "%d/%m/%Y %H:%M:%S")
                                                    else:
                                                        dateTenor = reportDict['dateTenor']
                                                else:
                                                    dateTenor = ''
                                                if 'other' in reportDict.keys():
                                                    other = reportDict['other']
                                                else:
                                                    other = ''
                                                dataFrameData.append(
                                                    [author, document, chapter, section, subsection, paragraph, item,
                                                     datePublicationObject, dateEffectiveObject, fixing, variantValue,
                                                     variantValueType, dateTenor, other])
    return pandas.DataFrame(dataFrameData,
                            columns=['Author', 'Document', 'Chapter', 'Section', 'Subsection', 'Paragraph', 'Item',
                                     'Date Publication', 'Date Effective', 'Fixing', 'Value', 'Value Type',
                                     'Date Tenor', 'Other'])

def parseRawReports(reportsJson):
    loadedResponse = json.loads(reportsJson)
    reports = []
    if isinstance(loadedResponse, dict):
        for key, item in loadedResponse.items():
            if '.json' in key:
                reportsRaw = json.loads(item)
                for reportRaw in reportsRaw:
                    report = Message.Report.fromJson(reportRaw)
                    reports.append(report)
    elif isinstance(loadedResponse, list):
        for reportRaw in loadedResponse:
            report = Message.Report.fromJson(reportRaw)
            reports.append(report)
    return reports



"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import pysftp
import os
import pandas
import datetime
from ..cfg import Configuration
from .. import Dates

def getConnection(host='', user='', password='', port=22):
    try:
        cnopts = pysftp.CnOpts()
    except:
        cnopts = pysftp.CnOpts(knownhosts='known_hosts')
    cnopts.hostkeys = None
    if host == '':
        host = Configuration.get('BFI_FTP_HOST')
    if user == '':
        user = Configuration.get('BFI_FTP_USER')
    if password == '':
        password = Configuration.get('BFI_FTP_PASSWORD')
    try:
        sftp = pysftp.Connection(host, username=user, password=password, cnopts=cnopts, port=port)
        return sftp
    except Exception as e:
        return 'Credentials (host, user, password, port, filepath) not valid. (Exception: {})'.format(e)

def downloadFile(filepath, host='', user='', password='', port=22, tmp=False):
    if filepath[0] == '/':
        filepath = filepath[1:]
    filename = filepath.split('/')[-1]
    if tmp:
        filename = os.path.join('/tmp', filename)
    sftp = getConnection(host, user, password, port)
    if sftp.isfile(filepath):
        sftp.get(filepath, filename)
    else:
        return "File doesn't exist."
    sftp.close()

def getFileAsString(filepath, host='', user='', password='', port=22, tmp=False):
    if filepath[0] == '/':
        filepath = filepath[1:]
    filename = filepath.split('/')[-1]
    downloadFile(filepath, host, user, password, port, tmp)
    try:
        file = open(filename, "r").read()
    except:
        file = open(filename, "r", encoding="utf8").read()
    os.remove(filename)
    return file

def getCSVFileAsPandas(filepath, separator=',', decimal='.', header="infer", host='', user='', password='', port=22, tmp=False):
    if filepath[0] == '/':
        filepath = filepath[1:]
    filename = filepath.split('/')[-1]
    if tmp:
        filename = os.path.join('/tmp', filename)
    sftp = getConnection(host, user, password, port)
    if sftp.isfile(filepath):
        sftp.get(filepath, filename)
    else:
        return "File doesn't exist."
    sftp.close()
    file = pandas.read_csv(filename, sep=separator, index_col=False, decimal=decimal, header=header)
    os.remove(filename)
    return file

def getExcelFileAsPandas(filepath, host='', user='', password='', port=22, tmp=False):
    if filepath[0] == '/':
        filepath = filepath[1:]
    filename = filepath.split('/')[-1]
    if tmp:
        filename = os.path.join('/tmp', filename)
    sftp = getConnection(host, user, password, port)
    if sftp.isfile(filepath):
        sftp.get(filepath, filename)
    else:
        return "File doesn't exist."
    sftp.close()
    file = pandas.read_excel(filename)
    os.remove(filename)
    return file

def uploadPandasAsCSV(dataFrame, filepath, host='', user='', password='', port=22, tmp=False, header=True):
    if filepath[0] == '/':
        filepath = filepath[1:]
    filename = filepath.split('/')[-1]
    if tmp:
        filename = os.path.join('/tmp', filename)
    dataFrame.to_csv(filename, index=False, header=header)
    sftp = getConnection(host, user, password, port)
    sftp.put(filename, filepath)
    sftp.close()
    os.remove(filename)

def uploadFileFromString(stringFile, filepath, host='', user='', password='', port=22, tmp=False):
    if filepath[0] == '/':
        filepath = filepath[1:]
    filename = filepath.split('/')[-1]
    if tmp:
        filename = os.path.join('/tmp', filename)
    with open(filename, 'w', encoding="utf-8") as outFile:
        outFile.write(stringFile)
    sftp = getConnection(host, user, password, port)
    sftp.put(filename, filepath)
    sftp.close()
    os.remove(filename)

def uploadFile(localFilepath, remoteFilepath, host='', user='', password='', port=22, removeFile=False):
    if remoteFilepath[0] == '/':
        remoteFilepath = remoteFilepath[1:]
    sftp = getConnection(host, user, password, port)
    sftp.put(localFilepath, remoteFilepath)
    sftp.close()
    if removeFile:
        os.remove(localFilepath)

def getListOfFilesInDir(dirpath, host='', user='', password='', port=22):
    if dirpath[0] == '/':
        dirpath = dirpath[1:]
    sftp = getConnection(host, user, password, port)
    return [filename for filename in sftp.listdir(dirpath)]

def deleteFile(filepath, host='', user='', password='', port=22):
    if filepath[0] == '/':
        filepath = filepath[1:]
    sftp = getConnection(host, user, password, port)
    sftp.remove(filepath)
    sftp.close()

def deleteMultipleFiles(directoryPath, filenames, host='', user='', password='', port=22):
    if directoryPath[0] == '/':
        directoryPath = directoryPath[1:]
    sftp = getConnection(host, user, password, port)
    for filename in filenames:
        filepath = '{}/{}'.format(directoryPath, filename)
        try:
            sftp.remove(filepath)
        except Exception as e:
            print(e)
    sftp.close()

def cleanDirectory(directoryPath, host='', user='', password='', port=22):
    if directoryPath[0] == '/':
        directoryPath = directoryPath[1:]
    sftp = getConnection(host, user, password, port)
    try:
        filesInRemoteArtifacts = sftp.listdir(directoryPath)
        for file in filesInRemoteArtifacts:
            filePath = '{}/{}'.format(directoryPath, file)
            sftp.remove(filePath)
        sftp.close()
    except:
        print("The directory is empty!")

def cleanTMPDirectory(retentionPeriod=2):
    host = ''
    user = ''
    password = ''
    port = 22
    directoryPath = 'backup/tmp'
    sftp = getConnection(host, user, password, port)
    try:
        filesInRemoteArtifacts = sftp.listdir(directoryPath)
        for file in filesInRemoteArtifacts:
            dateFile = Dates.getDateTimeFromFilenameTimestamp(file)
            if (datetime.datetime.now() - dateFile).days > retentionPeriod:
                filePath = '{}/{}'.format(directoryPath, file)
                sftp.remove(filePath)
        sftp.close()
    except:
        print("The tmp directory is empty!")


if __name__ == '__main__':
    # file = uploadFileFromString('Hey,Ho', 'pacifico/unavailableTEST.csv')
    # print(file)
    #import datetime
    #today = datetime.datetime.now()
    #file = file.append({'Ticker': 'AAPL', 'Date': today, 'Author': 'Benjamin2'}, ignore_index=True)
    #uploadPandasAsCSV(file, 'ec2-54-203-94-208.us-west-2.compute.amazonaws.com', 'bcs', 'bolsa2020', 22, 'pacifico/unavailable.csv')
    ftpDir = 'backup/tmp'
    cleanDirectory(ftpDir)


import typing as typ
import pandas as pd
import altair as alt

import pacifico_devel.util.reportTemplate.src.core.Plot.PlotAltair.PlotAltair as plta
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl
import pacifico_devel.util.reportTemplate.src.util.PacificoPlotColors as pc

import pacifico_devel.util.lexikon.src.util.Enumerations as lenm
import pacifico_devel.util.lexikon.src.core.Translator as trl


## IDEA: collapse languageInput and languageOutput into one.


class PlotAltairLine(plta.PlotAltair):
    def __init__(self,
                 df: pd.DataFrame,
                 languageInput: lenm.Language,
                 languageOutput: lenm.Language,
                 shape: typ.Tuple[float, float],
                 basePlotResolution: int,
                 style: typ.Optional[stl.Style] = None,
                 title: str = None):
        super(PlotAltairLine, self).__init__(df=df,
                                             shape=shape,
                                             languageInput=languageInput,
                                             languageOutput=languageOutput,
                                             basePlotResolution=basePlotResolution,
                                             style=style)
        self._title = title

    def _makeChart(self) -> alt.Chart:
        indexName = self.df.index.name
        data = self.df[[colName
                        for colName in self.df.columns
                        if pd.api.types.is_numeric_dtype(self.df[colName])]]
        data = data.reset_index()

        varName = trl.Translator.translate(text="series",
                                           languageInput=self._languageInput,
                                           languageOutput=self._languageOutput)
        valueName = trl.Translator.translate(text="value",
                                             languageInput=self._languageInput,
                                             languageOutput=self._languageOutput)

        data = data.melt(id_vars=data.columns[0],
                         var_name=varName,
                         value_name=valueName)
        colorScale = alt.Scale(domain=data[varName].unique(),
                               range=[c.getCodeString()
                                      for c in pc.PacificoPlotColors]
                               )
        chart = alt.Chart(data).mark_line().encode(x=indexName,
                                                   y=valueName,
                                                   color=alt.Color(varName, scale=colorScale))
        chart = chart.properties(title=self._title,
                                 width=self._shape[0],
                                 height=self._shape[1])
        return chart



import abc
import pathlib as pth
import bokeh.io as bio
import pandas as pd
import typing as typ
import bokeh.plotting as bpl
import bokeh.models as bm
from bokeh.plotting import output_file, show

import pacifico_devel.util.reportTemplate.src.util.CONFIG as CFG

import pacifico_devel.util.reportTemplate.src.core.Plot.Plot as bplt


class PlotBokeh(bplt.Plot, metaclass=abc.ABCMeta):
    """
    Abstract class for all necessary plots made with Bokeh Backend.
    """
    _BOKEH_THEME_PATH = CFG.PATH_BOKEH_THEME

    # @property
    # def _theme_path(self):
    #     return self._BOKEH_THEME_PATH

    def saveImage(self) -> None:
        """


        Args:
            dfTranslated:

        Returns:

        """
        output_file("Plot.html")  # Was Necessary.
        chart = self._makeChart()
        print(self.pathImage)
        bio.export_png(chart,
                       filename=self.pathImage)
        print(self.pathImage)

    @abc.abstractmethod
    def _makeChart(self) -> typ.Union[bpl.Figure, bm.DataTable]:
        """
        Makes chart.

        Args:
            dfTranslated:

        Returns:

        """
        pass

    def df_to_excel(self, writer, tag):
        """
        saves DataFrame to excel
        Returns:

        """
        self.df.to_excel(writer, sheet_name=tag)



"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import os
import sys
import paramiko
import time
from ..aws import EC2, Enumerations
from ..cfg import Configuration


def connectSSH(ip_address, pemFilePath):
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    privkey = paramiko.RSAKey.from_private_key_file(pemFilePath)  # TODO: Fix Auth, works for paramiko==2.7.2, but not for new versions.
    try:
        client.connect(hostname=ip_address, username='centos', pkey=privkey)
        return client
    except (Exception, paramiko.BadHostKeyException, paramiko.AuthenticationException, paramiko.SSHException) as exception:
        print(exception)
        return __buildErrorOutput(str(exception))


def connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    ip_address = EC2.getInstancePublicIPAddress(instanceId, awsAccessKey, awsSecretKey)
    client = connectSSH(ip_address, pemFilePath)
    return client


def __buildErrorOutput(message):
    status = -1
    error = message
    out = 'error'
    return status, error, out


def __executeCommandWithoutWaitingForCompletionUsingSudo(command, client):
    if not isinstance(client, tuple):
        client.get_transport()
        client.exec_command(command, get_pty=True)
        status = 0
        error = ''
        out = 'Sent without waiting for completion and with sudo permissions'
        return status, error, out
    else:
        return client


def __executeCommandWithoutWaitingForCompletion(command, client):
    if not isinstance(client, tuple):
        client.exec_command(command)
        status = 0
        error = ''
        out = 'Sent without waiting for completion'
        return status, error, out
    else:
        return client


def __executeCommandWaitingForCompletion(command, client, timeout=None, debug=False):
    if not isinstance(client, tuple):
        if debug:
            print(f'{command} started...')
        out = f'{command}\n\n'
        error = f'{command}\n\n'
        stdin, stdout, stderr = client.exec_command(command, timeout=timeout, get_pty=True)
        stdin.flush()
        # https://stackoverflow.com/questions/31834743/get-output-from-a-paramiko-ssh-exec-command-continuously
        for line in iter(stdout.readline, ""):
            if debug:
                print(line, end="")
            out += line
            stdout.flush()
            stderr.flush()
        for errorLine in iter(stderr.readline, ""):
            if debug:
                print(errorLine, end="")
            error += errorLine
            stdout.flush()
            stderr.flush()
        if debug:
            print(f'{command} finished.')
        status = stdout.channel.recv_exit_status()
        # Waiting for full response (buffering for big outputs that take a while issue)
        # error = stderr.read().decode('UTF-8')
        # out = stdout.read().decode('UTF-8')
        if debug:
            print(f'{command} status: {status}...')
        return status, error, out
    else:
        return client

def __executeCommand(command, client, timeout=None):
    return __executeCommandWaitingForCompletion(command, client, timeout)

def executeCommand(command, client, instanceId, timeout=None):
    if command[:5] == './bfi':
        command = EC2.__getRoot() + command
        logFilepath = '/bfi/log/bfi_log_{}.txt'.format(instanceId)
        # if waitForCompletion == Enumerations.WaitForCompletion.NotWait or waitForCompletion == Enumerations.WaitForCompletion.StartStop:
        #     # 2>&1: https://unix.stackexchange.com/questions/45913/is-there-a-way-to-redirect-nohup-output-to-a-log-file-other-than-nohup-out
        # command = command + ' > {} 2>&1'.format(logFilepath)
        #     if waitForCompletion == Enumerations.WaitForCompletion.StartStop:
        #         # shutdown:  # https://dev.to/aws/auto-stop-ec2-instances-when-they-finish-a-task-2f0i
        #         # sudo: https://www.youtube.com/watch?v=fVOFWehhc38
        #         command = "sudo -b sh -c '" + '{ su - centos -c "' + command + '"; shutdown now -h; ' + "}'"
        #     else:
        #         # nohup + &: https://docs.aws.amazon.com/es_es/codebuild/latest/userguide/build-env-ref-background-tasks.html
        #         # setsid: https://unix.stackexchange.com/questions/432258/issues-with-nohup-on-linux-instance-on-ec2-amazon-web-services
        #         command = 'setsid nohup {} &'.format(command)
        # else:
            # | tee: https://stackoverflow.com/questions/13591374/command-output-redirect-to-file-and-terminal
            # exitcode fix: https://stackoverflow.com/questions/6871859/piping-command-output-to-tee-but-also-save-exit-code-of-command
        command = 'set -o pipefail && {} 2>&1 | tee {}'.format(command, logFilepath)
        return __executeCommand(command, client, timeout)
    else:
        return __buildErrorOutput('Invalid command')


def executeCommandInInstance(command, instanceId, pemFilePath, awsAccessKey, awsSecretKey, timeout=None):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    return executeCommand(command, client, instanceId, timeout)


def copyFileFromInstance(filePathInstance, filepathLocal, instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    if not isinstance(client, tuple):
        ftp_client = client.open_sftp()
        try:
            ftp_client.get(filePathInstance, filepathLocal)
        except Exception as e:
            print(f"SSH: File {filePathInstance} couldn't be copied from instance {instanceId} ({str(e)})!")
        ftp_client.close()
        client.close()

def copyMultipleFilesFromInstance(instancePath, filenames, localPath, instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    if not isinstance(client, tuple):
        ftp_client = client.open_sftp()
        for filename in filenames:
            filePathInstance = os.path.join(instancePath, filename)
            filepathLocal = os.path.join(localPath, filename)
            try:
                ftp_client.get(filePathInstance, filepathLocal)
            except Exception as e:
                print(f"SSH: File {filePathInstance} couldn't be copied from instance {instanceId} ({str(e)})!")
        ftp_client.close()
        client.close()


def readFileFromInstance(filePathInstance, instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    if not isinstance(client, tuple):
        ftp_client = client.open_sftp()
        dataAsString = ftp_client.open(filePathInstance, mode='r').read().decode("utf-8")
        ftp_client.close()
        client.close()
        return dataAsString
    else:
        return str(client)


def writeFileToInstance(filePathInstance, filepathLocal, instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    if not isinstance(client, tuple):
        ftp_client = client.open_sftp()
        ftp_client.put(filepathLocal, filePathInstance)
        ftp_client.close()
        client.close()


def writeFileStringToInstance(stringData, filePath, instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    if not isinstance(client, tuple):
        ftp_client = client.open_sftp()
        file = ftp_client.file(filePath, "a")
        file.write(stringData)
        file.flush()
        ftp_client.close()
        client.close()


def deleteFileFromInstance(filePathInstance, instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    if not isinstance(client, tuple):
        ftp_client = client.open_sftp()
        ftp_client.remove(filePathInstance)
        ftp_client.close()
        client.close()

def renameFileFromInstance(filePathInstance, newFilePathInstance, instanceId, pemFilePath, awsAccessKey, awsSecretKey):
    client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    if not isinstance(client, tuple):
        ftp_client = client.open_sftp()
        ftp_client.rename(filePathInstance, newFilePathInstance)
        ftp_client.close()
        client.close()


def getRunningProcessesInInstance(client):
    command = 'ps -A'  # == 'ps -e'
    status, error, out = __executeCommand(command, client)
    return out


def isProcessRunningInInstance(processName, client):
    processesString = getRunningProcessesInInstance(client)
    if processName in processesString:
        return True
    else:
        return False


def isInstanceBusy(client):
    processName = 'bfi'
    return isProcessRunningInInstance(processName, client)


if __name__ == '__main__':
    import time

    awsAccessKey = Configuration.get('AWS_ACCESS_KEY')
    awsSecretKey = Configuration.get('AWS_SECRET_KEY')
    instanceId = Configuration.get('COMPUTER_2_INSTANCE_ID')  # Configuration.get('COMPUTER_STATIC_2_INSTANCE_ID')
    pemFilePath = 'computer_key.pem'
    # Test commands

    command = './bfi_api -datawarehouse -refresh'
    statusCode, error, out = executeCommandInInstance(command, instanceId, pemFilePath, awsAccessKey, awsSecretKey,
                                                      Enumerations.WaitForCompletion.StartStop)
    print('out:')
    print(out)
    print('error:')
    print(error)
    print('statusCode:', statusCode)
    '''
    while True:
        client = connectSSHByInstanceId(instanceId, pemFilePath, awsAccessKey, awsSecretKey)
        print(isInstanceBusy(client))
        time.sleep(30)

    # Test waitForCompletion
    command = './bfi_api -bcs -upload -parameters' #'./bfi_api -help'  # './bfi_api ls'
    timeStart = time.time()
    statusCode, error, out = executeCommandInInstance(command, instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    print(time.time() - timeStart)
    print(out)
    print(error)
    print(statusCode)
    print('-------------------------------')
    timeStart = time.time()
    statusCode, error, out = executeCommandInInstance(command, instanceId, pemFilePath, awsAccessKey, awsSecretKey, False)
    print(time.time() - timeStart)
    print(out)
    print(error)
    print(statusCode)
    #writeFileStringToInstance('12345', EC2.__getRoot() + 'test.txt', instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    #writeFileToInstance(EC2.__getRoot() + 'computer_key.pem', 'tmp/computer_key.pem', instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    #copyFileFromInstance(EC2.__getRoot() + 'test.txt', 'tmp/test.txt', instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    #file = readFileFromInstance(EC2.__getRoot() + 'test.txt', instanceId, pemFilePath, awsAccessKey, awsSecretKey)
    #print(file)
    '''


try:
    import psycopg2
except Exception as e:
    print(f'Error importing psycopg2 ({str(e)})')

class Connection:

    @staticmethod
    def connect(user="pacifico", password="PACIFICO", host="pacifico.cx4dwi9ydr2o.us-west-2.rds.amazonaws.com", port="5432", database="pacifico"):
        connection = psycopg2.connect(user=user,
                                      password=password,
                                      host=host,
                                      port=port,
                                      database=database)
        return connection


import typing as typ
import pandas as pd
import altair as alt
import bokeh.plotting as bpl
import bokeh.themes as bt
import bokeh.models as bm
from bokeh.plotting import output_file, show

import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokeh as pltb
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl
import pacifico_devel.util.reportTemplate.src.util.PacificoPlotColors as pc

import pacifico_devel.util.lexikon.src.util.Enumerations as lenm

import pacifico_devel.util.lexikon.src.core.Translator as trl


## IDEA: collapse languageInput and languageOutput into one.


class PlotBokehLineTwoYAxis(pltb.PlotBokeh):
    def __init__(self,
                 df: pd.DataFrame,
                 shape: typ.Tuple[float, float],
                 languageInput: lenm.Language,
                 languageOutput: lenm.Language,
                 basePlotResolution: int,
                 style: typ.Optional[stl.Style] = None,
                 title: str = None):
        super(PlotBokehLineTwoYAxis, self).__init__(df=df, shape=shape, languageInput=languageInput,
                                                    languageOutput=languageOutput,
                                                    basePlotResolution=basePlotResolution,
                                                    style=style)
        self._title = title

    def _makeChart(self) -> bpl.Figure:
        # output_file("line.html")
        chart = bpl.figure(width=self._shape[0],
                           height=self._shape[1],
                           toolbar_location=None)
        bpl.curdoc().theme = bt.Theme(filename=self._BOKEH_THEME_PATH)
        chart.title = self._title
        if pd.api.types.is_datetime64_any_dtype(self.df.index):
            if '-' in self._DATE_TIME_FORMAT:
                self._DATE_TIME_FORMAT = self._DATE_TIME_FORMAT.replace("-%d-", '-')
            if '/' in self._DATE_TIME_FORMAT:
                self._DATE_TIME_FORMAT = self._DATE_TIME_FORMAT.replace("%d/", '')
            chart.xaxis[0].formatter = bm.DatetimeTickFormatter(years=self._DATE_TIME_FORMAT,
                                                                months=self._DATE_TIME_FORMAT)
            # days=self._DATE_TIME_FORMAT)
        for index, (colName, color) in enumerate(zip(self.df.columns,
                                                   pc.PacificoPlotColors)):
            if pd.api.types.is_numeric_dtype(self.df[colName]):
                data = self.df[colName]
                if index == len(self.df.columns) - 1:
                    chart.extra_y_ranges = {"y2": bm.Range1d(start=min(data) * 1.3, end=max(data) * 1.3)}
                    chart.add_layout(bm.LinearAxis(y_range_name="y2"), 'right')
                    chart.line(x=data.index,
                               y=data,
                               legend_label=colName,
                               color=color.getCodeString(),
                               # source=data,
                               line_width=2,
                               y_range_name="y2")
                else:
                    chart.line(x=data.index,
                               y=data,
                               legend_label=colName,
                               color=color.getCodeString(),
                               # source=data,
                               line_width=2)

        self.style(chart)
        self.legendStyle(chart, 'horizontal', 'below')
        # show(chart)
        return chart

    def legendStyle(self, plot, orientation: str = 'vertical', layout: str = 'right'):
        new_legend = plot.legend[0]
        plot.add_layout(new_legend, layout)  # 'below'
        plot.legend.orientation = orientation  # 'horizontal'
        plot.legend.location = "center"
        plot.legend.border_line_alpha = 0
        plot.legend.background_fill_color = None
        plot.xgrid.grid_line_color = None
        plot.legend.click_policy = "hide"
        plot.legend.label_text_font = "klartext mono"
        plot.legend.label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        return plot

    def style(self, plot):
        # TITLE
        plot.title.text_color = pc.PacificoPlotColors.BLUE.getCodeString()
        plot.title.text_font_size = "20px"
        plot.title.text_font = "klartext mono"
        # plot.title.text_font_style = "italic"
        # LABELS
        plot.axis.axis_label_text_font = "klartext mono"
        plot.axis.axis_label_text_font_size = "15px"
        plot.axis.axis_label_text_font_style = 'normal'
        plot.axis.axis_label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        plot.axis.major_label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        # AXIS
        plot.xaxis.axis_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.axis.major_tick_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.xaxis.major_tick_line_width = 1
        plot.axis.minor_tick_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.yaxis.minor_tick_line_color = None
        plot.yaxis.axis_line_color = None
        # TICKS
        # plot.xaxis.major_label_orientation = np.pi / 2
        plot.xaxis.major_label_orientation = "vertical"
        # BACKGROUND
        plot.xgrid.grid_line_color = None
        plot.outline_line_color = None
        plot.yaxis[0].formatter = bm.NumeralTickFormatter(format="y{0%} ")
        plot.yaxis[1].formatter = bm.NumeralTickFormatter(format="y{0%} ")
        return plot


if __name__ == '__main__':
    from datetime import datetime
    from Applications.ApplicationReportChileUnemployment.src.core.Data import Data
    from pacifico_devel.util.lexikon.src.util.Enumerations import Language
    import chromedriver_autoinstaller

    chromedriver_autoinstaller.install()
    data = Data(datetime.now())
    data.load('KmK9FqgJ5h1pSNjh1vhha6ul30vXBOBiMVlmplXf', ['60703000-6', 'F049.CES.TAS.UCH3.65.T'])
    data.computeUnemployment()
    data = data.getData()
    # with pd.option_context('display.max_rows', None, 'display.max_columns', None):
    #    print(data[''])
    years = list(set(data['Año']))
    years.sort()
    series = []
    seriesNames = ['Fuerza de trabajo (Total)',
                   'Asalariados/as (Sector privado)',
                   'Asalariados/as (Sector público) /5']
    periods = 12

    serie = Data.percentChange(data[seriesNames], periods)  # data[criteria].dropna().pct_change(periods=12) * 100
    date = datetime(years[-4], max(serie.index).month, 1)
    series.append(serie[serie.index >= date])
    dataPlot = pd.concat(series, axis=1)

    p = PlotBokehLineTwoYAxis(df=dataPlot,
                              shape=(1, 1 / 4),
                              languageInput=Language.fromDeeplString("ES"),
                              languageOutput=Language.fromDeeplString("ES"),
                              basePlotResolution=1000,
                              style=None,
                              title='')._makeChart()
    show(p)



"""Copyright © 2022 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import email
import imaplib
import smtplib
from email.header import decode_header
from email.mime.text import MIMEText

class EmailServer:
    # https://humberto.io/blog/sending-and-receiving-emails-with-python/
    def __init__(self, sendServer, sendPort, receiveServer, receivePort):
        self.sendServer = sendServer
        self.sendPort = sendPort
        self.receiveServer = receiveServer
        self.receivePort = receivePort

    def getSendServer(self):
        return self.sendServer

    def getSendPort(self):
        return self.sendPort

    def getReceiveServer(self):
        return self.receiveServer

    def getReceivePort(self):
        return self.receivePort

    def sendEmail(self, userEmail, userPassword, destinationEmailList, subject, message, html=False):
        # Connect with email provider servers
        smtp_ssl_host = self.getSendServer()  # 'smtp.gmail.com'
        smtp_ssl_port = self.getSendPort()  # 465
        # Use username or email to log in
        username = userEmail  # 'origin@gmail.com'
        password = userPassword  # 'password'
        from_addr = userEmail  # 'origin@gmail.com'
        if not html:
            message = MIMEText(message)  # 'Hello World')
        else:  # https://stackoverflow.com/questions/44322806/sending-email-with-color-formatting-in-python
            message = MIMEText(message, 'html')
        message['subject'] = subject  # 'Hello'
        message['from'] = from_addr
        message['to'] = ', '.join(destinationEmailList)  # ['destiny@gmail.com']
        # We'll connect using SSL
        server = smtplib.SMTP_SSL(smtp_ssl_host, smtp_ssl_port)
        server.login(username, password)
        server.sendmail(from_addr, destinationEmailList, message.as_string())
        server.quit()
        print(f'Email with subject "{subject}" from "{userEmail}" to "{", ".join(destinationEmailList)}" was sent correctly!')

    def getInbox(self, userEmail, userPassword, amountOfEmails=None, latest=True, verbose=False):
        emails = []
        EMAIL = userEmail  # 'mymail@mail.com'
        PASSWORD = userPassword  # 'password'
        SERVER = self.getReceiveServer()  # 'imap.gmail.com'
        # Connect to the server and go to its inbox
        mail = imaplib.IMAP4_SSL(SERVER)
        mail.login(EMAIL, PASSWORD)
        # We choose the inbox but you can select others
        mail.select('inbox')
        # We'll search using the ALL criteria to retrieve every message inside the inbox it will return with its status
        # and a list of ids
        status, data = mail.search(None, 'ALL')
        # The list returned is a list of bytes separated by white spaces on this format: [b'1 2 3', b'4 5 6'] so, to
        # separate it first we create an empty list
        mail_ids = []
        # Then we go through the list splitting its blocks of bytes and appending to the mail_ids list
        for block in data:
            # the split function called without parameter transforms the text or bytes into a list using as separator
            # the white spaces:
            # b'1 2 3'.split() => [b'1', b'2', b'3']
            mail_ids += block.split()
        # Now for every id we'll fetch the email
        if latest:  # To get the latest emails first
            mail_ids.reverse()
        if amountOfEmails is not None:  # To retrieve a specific amount of emails
            mail_ids = mail_ids[:amountOfEmails]
        for i in mail_ids:
            # The fetch function fetch the email given its id and format that you want the message to be
            status, data = mail.fetch(i, '(RFC822)')
            # The content data at the '(RFC822)' format comes on a list with a tuple with header, content, and the
            # closing byte b')'
            for response_part in data:
                # So if its a tuple...
                if isinstance(response_part, tuple):
                    # We go for the content at its second element skipping the header at the first and the closing at
                    # the third
                    message = email.message_from_bytes(response_part[1])
                    # With the content we can extract the info about who sent the message and its subject
                    mail_from = message.get('from')
                    mail_to = message.get('to')
                    mail_cc = message.get('cc')
                    mail_datetime = message.get('date')
                    # Mail subject special parsing
                    mail_subject = message.get('subject')
                    result = decode_header(mail_subject)
                    mail_subject_items = []
                    for data, encoding in result:
                        if encoding is not None:
                            data = data.decode(encoding)
                        elif isinstance(data, bytes):
                            data = data.decode('utf-8')
                        mail_subject_items.append(data)
                    mail_subject = ' '.join(mail_subject_items)
                    # Then for the text we have a little more work to do because it can be in plain text or multipart
                    # if its not plain text we need to separate the message from its annexes to get the text
                    if message.is_multipart():
                        mail_content = []
                        # On multipart we have the text message and another things like annex, and html version of the
                        # message, in that case we loop through the email payload
                        for part in message.get_payload():
                            # if part.get_content_type() == 'text/plain':
                            content = (part.get_content_type(), part.get_payload())
                            mail_content.append(content)
                    else:
                        # If the message isn't multipart, just extract it
                        mail_content = message.get_payload()
                    # Let's show its result
                    if verbose:
                        print(f'From: {mail_from}')
                        print(f'To: {mail_to}')
                        print(f'CC: {mail_cc}')
                        print(f'Datetime: {mail_datetime}')
                        print(f'Subject: {mail_subject}')
                        print(f'Content: {mail_content}')
                        print('*' * 100)
                    # Let's return its result
                    emailDict = {'From': mail_from, 'To': mail_to, 'CC': mail_cc, 'Datetime': mail_datetime,
                                 'Subject': mail_subject, 'Content': mail_content}
                    emails.append(emailDict)
        return emails

    def deleteEmail(self, userEmail, userPassword, subject):
        EMAIL = userEmail  # 'mymail@mail.com'
        PASSWORD = userPassword  # 'password'
        SERVER = self.getReceiveServer()  # 'imap.gmail.com'
        # Connect to the server and go to its inbox
        imap = imaplib.IMAP4_SSL(SERVER)
        imap.login(EMAIL, PASSWORD)
        # We choose the inbox but you can select others
        imap.select('inbox')
        # We'll search using the ALL criteria to retrieve every message inside the inbox it will return with its status
        # and a list of ids
        # to get mails by subject
        status, messages = imap.search(None, f'SUBJECT "{subject}"')
        # convert messages to a list of email IDs
        messages = messages[0].split(b' ')
        for mail in messages:
            _, msg = imap.fetch(mail, "(RFC822)")
            # you can delete the for loop for performance if you have a long list of emails
            # because it is only for printing the SUBJECT of target email to delete
            for response in msg:
                if isinstance(response, tuple):
                    msg = email.message_from_bytes(response[1])
                    # decode the email subject
                    subject = decode_header(msg["Subject"])[0][0]
                    if isinstance(subject, bytes):
                        # if it's a bytes type, decode to str
                        subject = subject.decode()
                    print("Deleting", subject)
            # mark the mail as deleted
            imap.store(mail, "+FLAGS", "\\Deleted")
        # permanently remove mails that are marked as deleted
        # from the selected mailbox (in this case, INBOX)
        imap.expunge()
        # close the mailbox
        imap.close()
        # logout from the account
        imap.logout()

if __name__ == '__main__':
    sendServer = 'bfi.lat'
    sendPort = 465  # 587
    receiveServer = 'mail.bfi.lat'
    receivePort = 110
    server = EmailServer(sendServer, sendPort, receiveServer, receivePort)
    userEmail = 'backoffice@bfi.lat'
    userPassword = 'backoffice2022'
    emails = server.getInbox(userEmail, userPassword, amountOfEmails=100)
    for email in emails:
        print(email['From'])
    destinationEmailList = ['benyin96@gmail.com']
    subject = 'Test email from Python'
    message = 'Hello World!'
    server.sendEmail(userEmail, userPassword, destinationEmailList, subject, message)


"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

from enum import Enum

class UsagePlan(Enum):
    UsagePlan_Client = 0
    UsagePlan_Author = 1
    UsagePlan_Administrator = 2

class Action(Enum):
    Action_Enable = 1
    Action_Disable = -1

class Difficulty(Enum):
    Difficulty_Low = -100
    Difficulty_Standard = 0
    Difficulty_Static = 10
    Difficulty_High = 100
    Difficulty_Application = 50

class WaitForCompletion(Enum):
    Wait = 1
    NotWait = 0
    StartStop = -1
    Timeout_1_Minute = 60
    Timeout_2_Minutes = 120
    Timeout_3_Minutes = 180
    Timeout_4_Minutes = 240
    Timeout_5_Minutes = 300
    Timeout_6_Minutes = 360
    Timeout_7_Minutes = 420
    Timeout_8_Minutes = 480
    Timeout_9_Minutes = 540
    Timeout_10_Minutes = 600
    Timeout_11_Minutes = 660
    Timeout_12_Minutes = 720
    Timeout_13_Minutes = 780
    Timeout_14_Minutes = 840
    Timeout_15_Minutes = 900
    Timeout_20_Minutes = 1200

    def getString(self):
        return self.name

    def getValue(self):
        return self.value

    @staticmethod
    def fromValue(value):
        return WaitForCompletion(value)

    @staticmethod
    def fromString(string):
        for waitForCompletion in WaitForCompletion:
            if string == waitForCompletion.getString():
                return waitForCompletion

    def isTimeout(self):
        if self.getString().split('_')[0] == 'Timeout':
            return True
        else:
            return False


"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import datetime
import json
from . import Field
from api.src.util import Enumerations, Dates


class FieldValue(Field.Field):
    def __init__(self, datePublication, fixing, ticker, field, dateEffective, dateTenor, value, other=''):
        self.datePublication = datePublication
        self.fixing = fixing
        self.setTicker(ticker)
        self.setField(field)
        self.setDateEffective(dateEffective)
        self.setDateTenor(dateTenor)
        self.value = value
        self.other = other

    # fixing
    def _get_fixing(self):
        return self.__fixing
    def _set_fixing(self, value):
        if not isinstance(value, Enumerations.Fixing):
            raise TypeError("The fixing must be set to a class 'Enumerations.Fixing'.")
        self.__fixing = value
    fixing = property(_get_fixing, _set_fixing)

    # ticker
    def _get_ticker(self):
        return self.__ticker
    def _set_ticker(self, value):
        if not isinstance(value, str):
            raise TypeError("The ticker must be set to a class 'ioIdentificationMnemo.IdentificationMnemo'.")
        self.__ticker = value
    ticker = property(_get_ticker, _set_ticker)

    # field
    def _get_field(self):
        return self.__field
    def _set_field(self, value):
        if not isinstance(value, Enumerations.FieldType):
            raise TypeError("The field must be set to a class 'Enumerations.FieldType'.")
        self.__field = value
    field = property(_get_field, _set_field)

    # dateEffective
    def _get_dateEffective(self):
        return self.__dateEffective
    def _set_dateEffective(self, value):
        if not isinstance(value, datetime.datetime):
            raise TypeError("The dateEffective must be set to a class 'datetime.datetime'.")
        self.__dateEffective = value
    dateEffective = property(_get_dateEffective, _set_dateEffective)

    # value
    def _get_value(self):
        return self.__value
    def _set_value(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("The value must be set to a integer or float.")
        self.__value = value
    value = property(_get_value, _set_value)

    # other
    def _get_other(self):
        return self.__other
    def _set_other(self, value):
        if not isinstance(value, str):
            raise TypeError("The other must be set to a string.")
        self.__other = value
    other = property(_get_other, _set_other)

    def getDatePublication(self):
        return self.datePublication

    def getFixing(self):
        return self.fixing

    def getTicker(self):
        return self.ticker

    def setTicker(self, ticker):
        self.ticker = ticker

    def getField(self):
        return self.field

    def setField(self, field):
        if isinstance(field, Enumerations.FieldType):
            self.field = field
        else:
            self.field = Enumerations.FieldType.fromValueOrString(field)

    def getDateEffective(self):
        return self.dateEffective

    def setDateEffective(self, dateEffective):
        if isinstance(dateEffective, (datetime.datetime, datetime.date)):
            self.dateEffective = dateEffective
        else:
            if dateEffective != '':
                self.dateEffective = Dates.dateTimeFromString(dateEffective)
            else:
                if isinstance(self.getDatePublication(), str):
                    if Dates.isDateTimeFromDateTimeOrDateString(self.getDatePublication()):
                        self.dateEffective = Dates.dateTimeFromString(self.getDatePublication())
                    else:
                        datePublication = Dates.dateFromString(self.getDatePublication())
                        hour, minute, second = Enumerations.Fixing.getTime(self.getFixing())
                        self.dateEffective = Dates.createDateTimeFromValues(datePublication.year, datePublication.month, datePublication.day, hour, minute, second)
                else:
                    hour, minute, second = Enumerations.Fixing.getTime(self.getFixing())
                    self.dateEffective = Dates.createDateTimeFromValues(self.getDatePublication().year, self.getDatePublication().month, self.getDatePublication().day, hour, minute, second)

    def getDateTenor(self):
        return self.dateTenor

    def setDateTenor(self, dateTenor):
        if isinstance(dateTenor, datetime.datetime):
            self.dateTenor = dateTenor
        else:
            if dateTenor != '':
                self.dateTenor = Dates.dateTimeFromString(dateTenor)
            else:
                self.dateTenor = self.getDateEffective()

    def getValue(self):
        return self.value

    def getOther(self):
        return self.other

    def getBase(self):
        for base in self.__class__.__bases__:
            return base.__name__

    @staticmethod
    def __serializeAndClean(dictionary):
        delete = []
        for key, value in zip(dictionary.keys(), dictionary.values()):
            if key in ['fixing', 'datePublication']:
                delete.append(key)
            if value in [[], {}, None, '']:
                delete.append(key)
            elif isinstance(value, list):
                dictionary[key] = [item.getJson(True) for item in value]
            elif isinstance(value, Enumerations.Enum):
                dictionary[key] = value.value
            elif isinstance(value, (datetime.date, datetime.datetime)):
                dictionary[key] = Dates.getDateOrDateTimeString(value)
            else:
                dictionary[key] = value
        for key in delete:
            del dictionary[key]
        return dictionary

    def getJson(self):
        # Get all attributes as a dictionary
        data = self.__dict__.copy()
        # Delete object private prefix from attribute names
        oldKeys = list(data.keys())
        for key in oldKeys:
            newKey = key.replace('_' + self.__class__.__name__ + '__', '')
            newKey = newKey.replace('_' + self.getBase() + '__', '')
            data[newKey] = data.pop(key)
        # Serialize object within the attribute and delete empty attributes
        data = self.__serializeAndClean(data)
        # jsonize data
        dataJson = json.dumps(data)
        return dataJson


import typing as typ
import pandas as pd
import bokeh.plotting as bpl
import bokeh.themes as bt
from bokeh.plotting import output_file

import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokeh as pltb
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl
import pacifico_devel.util.reportTemplate.src.util.PacificoPlotColors as pc

import pacifico_devel.util.lexikon.src.util.Enumerations as lenm


class PlotBokehBar(pltb.PlotBokeh):
    def __init__(self,
                 df: pd.DataFrame,
                 shape: typ.Tuple[float, float],
                 languageInput: lenm.Language,
                 languageOutput: lenm.Language,
                 basePlotResolution: int,
                 style: typ.Optional[stl.Style] = None,
                 title: str = None):
        super(PlotBokehBar, self).__init__(df=df,
                                           shape=shape,
                                           languageInput=languageInput,
                                           languageOutput=languageOutput,
                                           basePlotResolution=basePlotResolution,
                                           style=style)
        self._title = title

    def _makeChart(self) -> bpl.Figure:
        categoriesColumn = self.df.columns[0]
        valuesColumn = self.df.columns[1]

        self.df["barchart_color"] = self.df[valuesColumn].apply(self._colorMap)
        #output_file("bar.html")
        chart = bpl.figure(width=self._shape[0],
                           height=self._shape[1],
                           y_range=self.df.sort_values(valuesColumn,
                                                       ascending=True)[categoriesColumn],
                           toolbar_location=None)
        bpl.curdoc().theme = bt.Theme(filename=self._BOKEH_THEME_PATH)
        chart.title = self._title

        chart.hbar(y=categoriesColumn,
                   left=0,
                   right=valuesColumn,
                   height=0.5,
                   color="barchart_color",
                   source=self.df)

        return chart

    @staticmethod
    def _colorMap(x: float):
        if x > 0:
            return pc.PacificoPlotColors.BLUE.getCodeString()
        return pc.PacificoPlotColors.PINK.getCodeString()



from enum import Enum, auto
from typing import Tuple, Optional
import seaborn as sns


class PacificoPlotColors(Enum):
    """
    Class for geting colors of pacifico plots
    """
    PINK = auto()
    BLUE = auto()
    TURQUOISE = auto()
    METALIC_BLUE = auto()
    DARKER_GREY = auto()
    DARK_GREY = auto()
    YELLOW = auto()
    ORANGE = auto()
    GREEN = auto()
    LIGHT_GREY = auto()
    LIGHTER_GREY = auto()
    PURPLE = auto()

    def getCodeString(self) -> str or list:
        """
        Get CodeString for a given color.

        Returns:

        """
        if self is PacificoPlotColors.BLUE:
            return "#330099"
        if self is PacificoPlotColors.PINK:
            return "#FF2E66"
        if self == PacificoPlotColors.METALIC_BLUE:
            return "#27576e"
        if self is PacificoPlotColors.TURQUOISE:
            return "#00D3B1"
        if self == PacificoPlotColors.DARKER_GREY:
            return "#505050"
        if self is PacificoPlotColors.DARK_GREY:
            return "#636363"
        if self is PacificoPlotColors.YELLOW:
            return "#FFD700"
        if self is PacificoPlotColors.ORANGE:
            return "#FF663D"
        if self is PacificoPlotColors.GREEN:
            return "#2BC655"
        if self is PacificoPlotColors.LIGHT_GREY:
            return "#D2D2D2"
        if self == PacificoPlotColors.LIGHTER_GREY:
            return "#dcdcdc"
        if self == PacificoPlotColors.PURPLE:
            return "#505050"

    def getRgbTuple(self, normalize=False) -> Tuple[int]:
        """
        get RGB float tuple.

        Args:
            normalize:

        Returns:

        """
        if self is PacificoPlotColors.PINK:
            out = 255, 46, 102
        elif self is PacificoPlotColors.BLUE:
            out = 39, 19, 129
        elif self is PacificoPlotColors.TURQUOISE:
            out = 0, 21, 177
        elif self is PacificoPlotColors.DARK_GREY:
            out = 99, 99, 99
        elif self is PacificoPlotColors.YELLOW:
            out = 255, 215, 0
        elif self is PacificoPlotColors.ORANGE:
            out = 43, 298, 85
        elif self is PacificoPlotColors.GREEN:
            out = 43, 298, 85
        elif self is PacificoPlotColors.LIGHT_GREY:
            out = 210, 210, 210
        else:
            raise ValueError(f"Invalid PacificoPlotColors: {self}")

        if normalize:
            out = (score / 255 for score in out)
        return out

    @classmethod
    def setPallete(cls,
                   lenPallete: Optional[int] = None,
                   nColors: int = 3,
                   returnPallete=False) -> None:
        """
        Set pallete for seaborn plot.

        Args:
            lenPallete:
            nColors:
            returnPallete:

        Returns:

        """
        if lenPallete is None:
            lenPallete = nColors
        colorStrings = [color.getCodeString() for color in cls]
        pallete = sns.blend_palette(colorStrings[:nColors],
                                    n_colors=lenPallete)
        sns.set_palette(pallete)
        if returnPallete:
            return pallete


if __name__ == '__main__':
    print(PacificoPlotColors.DARK_GREY.getCodeString())


import typing as typ
import pandas as pd
import altair as alt
import bokeh.plotting as bpl
import bokeh.themes as bt
import bokeh.models as bm
from bokeh.models import NumeralTickFormatter
from bokeh.plotting import output_file, show

import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokeh as pltb
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl
import pacifico_devel.util.reportTemplate.src.util.PacificoPlotColors as pc

import pacifico_devel.util.lexikon.src.util.Enumerations as lenm

import pacifico_devel.util.lexikon.src.core.Translator as trl


## IDEA: collapse languageInput and languageOutput into one.


class PlotBokehLineAndPointed(pltb.PlotBokeh):
    def __init__(self,
                 df: pd.DataFrame,
                 shape: typ.Tuple[float, float],
                 languageInput: lenm.Language,
                 languageOutput: lenm.Language,
                 basePlotResolution: int,
                 style: typ.Optional[stl.Style] = None,
                 title: str = None):
        super(PlotBokehLineAndPointed, self).__init__(df=df, shape=shape, languageInput=languageInput,
                                                      languageOutput=languageOutput,
                                                      basePlotResolution=basePlotResolution,
                                                      style=style)
        self._title = title

    def _makeChart(self) -> bpl.Figure:
        # output_file("line.html")
        chart = bpl.figure(width=self._shape[0],
                           height=self._shape[1],
                           toolbar_location=None)
        bpl.curdoc().theme = bt.Theme(filename=self._BOKEH_THEME_PATH)
        chart.title = self._title
        if pd.api.types.is_datetime64_any_dtype(self.df.index):
            if '-' in self._DATE_TIME_FORMAT:
                self._DATE_TIME_FORMAT = self._DATE_TIME_FORMAT.replace("-%d-", '-')
            if '/' in self._DATE_TIME_FORMAT:
                self._DATE_TIME_FORMAT = self._DATE_TIME_FORMAT.replace("%d/", '')
            chart.xaxis[0].formatter = bm.DatetimeTickFormatter(years=self._DATE_TIME_FORMAT,
                                                                months=self._DATE_TIME_FORMAT)
            # days=self._DATE_TIME_FORMAT)
        for colName, color in zip(self.df.columns,
                                  pc.PacificoPlotColors):
            if pd.api.types.is_numeric_dtype(self.df[colName].astype('float')):
                data = self.df[colName].astype('float')
                chart.line(x=data.index,
                           y=data,
                           legend_label=colName,
                           color=color.getCodeString(),
                           # source=data,
                           line_width=2)
                chart.circle(x=data.index,
                             y=data,
                             size=4,
                             fill_color=color.getCodeString(),
                             line_color=color.getCodeString(),
                             # source=data,
                             line_width=1)
        self.style(chart)
        self.legendStyle(chart, 'horizontal', 'below')
        # show(chart)
        return chart

    def legendStyle(self, plot, orientation: str = 'vertical', layout: str = 'right'):
        new_legend = plot.legend[0]
        plot.add_layout(new_legend, layout)  # 'below'
        plot.legend.orientation = orientation  # 'horizontal'
        plot.legend.location = "center"
        plot.legend.border_line_alpha = 0
        plot.legend.background_fill_color = None
        plot.xgrid.grid_line_color = None
        plot.legend.click_policy = "hide"
        plot.legend.label_text_font = "klartext mono"
        plot.legend.label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        return plot

    def style(self, plot):
        # TITLE
        plot.title.text_color = pc.PacificoPlotColors.BLUE.getCodeString()
        plot.title.text_font_size = "20px"
        plot.title.text_font = "klartext mono"
        # plot.title.text_font_style = "italic"
        # LABELS
        plot.axis.axis_label_text_font = "klartext mono"
        plot.axis.axis_label_text_font_size = "15px"
        plot.axis.axis_label_text_font_style = 'normal'
        plot.axis.axis_label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        plot.axis.major_label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        # AXIS
        plot.xaxis.axis_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.axis.major_tick_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.xaxis.major_tick_line_width = 1
        plot.axis.minor_tick_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.yaxis.minor_tick_line_color = None
        plot.yaxis.axis_line_color = None
        # TICKS
        # plot.xaxis.major_label_orientation = np.pi / 2
        plot.xaxis.major_label_orientation = "vertical"
        plot.yaxis[0].formatter = NumeralTickFormatter(format="y{0%} ")
        # BACKGROUND
        plot.xgrid.grid_line_color = None
        plot.outline_line_color = None

        return plot


if __name__ == '__main__':
    from datetime import datetime
    from Applications.ApplicationReportChileUnemployment.src.core.Data import Data
    from pacifico_devel.util.lexikon.src.util.Enumerations import Language
    import chromedriver_autoinstaller

    chromedriver_autoinstaller.install()
    data = Data(datetime.now())
    data.load('KmK9FqgJ5h1pSNjh1vhha6ul30vXBOBiMVlmplXf', ['60703000-6', 'F049.CES.TAS.UCH3.65.T'])
    data.computeUnemployment()
    data = data.getData()
    # with pd.option_context('display.max_rows', None, 'display.max_columns', None):
    #    print(data[''])
    years = list(set(data['Año']))
    years.sort()
    series = []
    seriesNames = ['Fuerza de trabajo (Total)', 'Población ocupada (Total)']
    periods = 12

    serie = Data.percentChange(data[seriesNames], periods)  # data[criteria].dropna().pct_change(periods=12) * 100
    date = datetime(years[-4], max(serie.index).month, 1)
    series.append(serie[serie.index >= date])
    dataPlot = pd.concat(series, axis=1)

    p = PlotBokehLineAndPointed(df=dataPlot,
                                shape=(1, 1 / 4),
                                languageInput=Language.fromDeeplString("ES"),
                                languageOutput=Language.fromDeeplString("ES"),
                                basePlotResolution=1000,
                                style=None,
                                title='')._makeChart()
    show(p)



"""Copyright © 2020 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import boto3
import os
from . import Enumerations, Configuration, Connection


class AmazonWebServices:

    @staticmethod
    def getToken(tokenId, awsAccessKey, awsSecretKey):
        client = boto3.client('apigateway', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey,
                              region_name='us-west-2')
        response = client.get_api_key(apiKey=tokenId, includeValue=True)
        token = response['value']
        return token

    @staticmethod
    def getName(tokenId, awsAccessKey, awsSecretKey):
        client = boto3.client('apigateway', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey,
                              region_name='us-west-2')
        response = client.get_api_key(apiKey=tokenId, includeValue=True)
        token = response['name']
        return token

    @staticmethod
    def createToken(name, awsAccessKey, awsSecretKey):
        client = boto3.client('apigateway', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey,
                              region_name='us-west-2')
        response = client.create_api_key(name=name, enabled=True)
        return response

    @staticmethod
    def __applySingleAction(tokenId, usagePlan, action, awsAccessKey, awsSecretKey, connection):
        if not isinstance(usagePlan, Enumerations.UsagePlan):
            raise TypeError("The usagePlan must be set to a class 'Enumerations.UsagePlan'.")
        if not isinstance(action, Enumerations.Action):
            raise TypeError("The action must be set to a class 'Enumerations.Action'.")
        client = boto3.client('apigateway', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey,
                              region_name='us-west-2')
        try:
            if action == Enumerations.Action.Action_Enable:
                if usagePlan == Enumerations.UsagePlan.UsagePlan_Administrator:
                    response = client.create_usage_plan_key(
                        usagePlanId=Configuration.Configuration.get('AWS_UsagePlanAdminID', connection),
                        keyId=tokenId,
                        keyType='API_KEY'
                    )
                elif usagePlan == Enumerations.UsagePlan.UsagePlan_Author:
                    response = client.create_usage_plan_key(
                        usagePlanId=Configuration.Configuration.get('AWS_UsagePlanAuthorID', connection),
                        keyId=tokenId,
                        keyType='API_KEY'
                    )
                elif usagePlan == Enumerations.UsagePlan.UsagePlan_Client:
                    response = client.create_usage_plan_key(
                        usagePlanId=Configuration.Configuration.get('AWS_UsagePlanUserID', connection),
                        keyId=tokenId,
                        keyType='API_KEY'
                    )
            elif action == Enumerations.Action.Action_Disable:
                if usagePlan == Enumerations.UsagePlan.UsagePlan_Administrator:
                    response = client.delete_usage_plan_key(
                        usagePlanId=Configuration.Configuration.get('AWS_UsagePlanAdminID', connection),
                        keyId=tokenId
                    )
                elif usagePlan == Enumerations.UsagePlan.UsagePlan_Author:
                    response = client.delete_usage_plan_key(
                        usagePlanId=Configuration.Configuration.get('AWS_UsagePlanAuthorID', connection),
                        keyId=tokenId
                    )
                elif usagePlan == Enumerations.UsagePlan.UsagePlan_Client:
                    response = client.delete_usage_plan_key(
                        usagePlanId=Configuration.Configuration.get('AWS_UsagePlanUserID', connection),
                        keyId=tokenId
                    )
        except:
            #print('AmazonWebServices.apply : This action ( ' + action.serial + ' - ' + usagePlan.serial + ' ) has already been applied to this token.')
            pass
        # print(response) # Uncomment to print response

    @staticmethod
    def apply(tokenId, usagePlan, action, awsAccessKey, awsSecretKey, connection):
        if action == Enumerations.Action.Action_Enable:
            if usagePlan == Enumerations.UsagePlan.UsagePlan_Administrator:
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Client, action, awsAccessKey, awsSecretKey, connection)
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Author, action, awsAccessKey, awsSecretKey, connection)
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Administrator, action, awsAccessKey, awsSecretKey, connection)
            elif usagePlan == Enumerations.UsagePlan.UsagePlan_Author:
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Client, action, awsAccessKey, awsSecretKey, connection)
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Author, action, awsAccessKey, awsSecretKey, connection)
            elif usagePlan == Enumerations.UsagePlan.UsagePlan_Client:
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Client, action, awsAccessKey, awsSecretKey, connection)
        elif action == Enumerations.Action.Action_Disable:
            if usagePlan == Enumerations.UsagePlan.UsagePlan_Administrator:
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Administrator, action, awsAccessKey, awsSecretKey, connection)
            elif usagePlan == Enumerations.UsagePlan.UsagePlan_Author:
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Author, action, awsAccessKey, awsSecretKey, connection)
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Administrator, action, awsAccessKey, awsSecretKey, connection)
            elif usagePlan == Enumerations.UsagePlan.UsagePlan_Client:
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Client, action, awsAccessKey, awsSecretKey, connection)
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Author, action, awsAccessKey, awsSecretKey, connection)
                AmazonWebServices.__applySingleAction(tokenId, Enumerations.UsagePlan.UsagePlan_Administrator, action, awsAccessKey, awsSecretKey, connection)

    @staticmethod
    def getUsagePlanKeys(usagePlan, awsAccessKey, awsSecretKey, connection):
        client = boto3.client('apigateway', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey,
                              region_name='us-west-2')
        if usagePlan == Enumerations.UsagePlan.UsagePlan_Client:
            plan = Configuration.Configuration.get("AWS_UsagePlanUserID", connection)
        elif usagePlan == Enumerations.UsagePlan.UsagePlan_Author:
            plan = Configuration.Configuration.get("AWS_UsagePlanAuthorID", connection)
        elif usagePlan == Enumerations.UsagePlan.UsagePlan_Administrator:
            plan = Configuration.Configuration.get("AWS_UsagePlanAdminID", connection)
        response = client.get_usage_plan_keys(
            usagePlanId=plan,
            #position='1',
            limit=500,
            #nameQuery='a'
        )
        return response['items']

    @staticmethod
    def getTokens(awsAccessKey, awsSecretKey):
        client = boto3.client('apigateway', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey,
                              region_name='us-west-2')
        response = client.get_api_keys(
            #position='string',
            limit=500,
            #nameQuery='string',
            #customerId='string',
            includeValues=True | False
        )
        return response['items']

    @staticmethod
    def getTokenId(token, awsAccessKey, awsSecretKey):
        tokens = AmazonWebServices.getTokens(awsAccessKey, awsSecretKey)
        tokenId = None
        for item in tokens:
            if item['value'] == token:
                tokenId = item['id']
        if tokenId != None:
            return tokenId
        else:
            raise KeyError('The token is not valid.')

    @staticmethod
    def isUsagePlanPermitted(token, usagePlan, awsAccessKey, awsSecretKey, connection):
        permitted = False
        tokenId = AmazonWebServices.getTokenId(token, awsAccessKey, awsSecretKey)
        keys = AmazonWebServices.getUsagePlanKeys(usagePlan, awsAccessKey, awsSecretKey, connection)
        for key in keys:
            if key['id'] == tokenId:
                permitted = True
        return permitted

    @staticmethod
    def getUrl(fileName, secondsItLasts, awsAccessKey, awsSecretKey, bucketName='', awsUrlAction=Enumerations.awsUrlAction.GET):
        if bucketName == '':
            bucketName = "media-pacifico"
        client = boto3.client('s3', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey, region_name='us-west-2')
        if awsUrlAction == Enumerations.awsUrlAction.PUT:
            response = client.generate_presigned_post(Bucket=bucketName, Key=fileName)  # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.generate_presigned_post
            # response = client.generate_presigned_url('put_object', Params={'Bucket': bucketName, 'Key': fileName}, ExpiresIn=secondsItLasts)  # https://stackoverflow.com/questions/52625812/boto3-generate-presigned-url-for-put-object-return-the-request-signature-we
        else:
            response = client.generate_presigned_url('get_object', Params={'Bucket': bucketName, 'Key': fileName}, ExpiresIn=secondsItLasts)
        return response

    @staticmethod
    def getUploadUrl(fileName, secondsItLasts, awsAccessKey, awsSecretKey, bucketName=''):
        if bucketName == '':
            bucketName = "media-pacifico"
        client = boto3.client('s3', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey, region_name='us-west-2')
        response = client.generate_presigned_url('put_object', Params={'Bucket': bucketName, 'Key': fileName}, ExpiresIn=secondsItLasts)
        return response

    @staticmethod
    def writeToBucket(stringData, filename, awsAccessKey, awsSecretKey, bucketFilename=''):
        if bucketFilename == '':
            bucketFilename = filename
        with open(filename, 'w') as outFile:
            outFile.write(stringData)
        client = boto3.client('s3', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey, region_name='us-west-2')
        bucketName = "media-pacifico"
        client.upload_file(filename, bucketName, bucketFilename)
        os.remove(filename)

    @staticmethod
    def downloadFromBucket(filepathBucket, awsAccessKey, awsSecretKey, filepathLocal='', bucketName="media-pacifico", read=True):
        client = boto3.client('s3', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey, region_name='us-west-2')
        if filepathLocal == '':
            filepathLocal = filepathBucket.split('/')[-1]
        client.download_file(bucketName, filepathBucket, filepathLocal)
        if read:
            with open(filepathLocal) as f:
                inputString = f.read()
            return inputString

    @staticmethod
    def getFileFromBucketAsString(filepathBucket, bucketName="media-pacifico", awsAccessKey='', awsSecretKey=''):
        client = boto3.client('s3', aws_access_key_id=awsAccessKey, aws_secret_access_key=awsSecretKey, region_name='us-west-2')
        data = client.get_object(Bucket=bucketName, Key=filepathBucket)['Body'].read()
        return data.decode("utf-8-sig")


if __name__=='__main__':
    from . import Enumerations

    connection = Connection.Connection.connect()
    token = 'kiHjqpmpIz8myYdoUpuXKa4rQ6IYaLva6wx9tFKy'
    usagePlan = Enumerations.UsagePlan.UsagePlan_Client
    id = AmazonWebServices.getTokenId(token, Configuration.Configuration.get('AWS_ACCESS_KEY', connection), Configuration.Configuration.get('AWS_SECRET_KEY', connection))
    usagePlanKeys = AmazonWebServices.getUsagePlanKeys(usagePlan, Configuration.Configuration.get('AWS_ACCESS_KEY', connection), Configuration.Configuration.get('AWS_SECRET_KEY', connection), connection)
    print(id)
    print(usagePlanKeys)
    permitted = AmazonWebServices.isUsagePlanPermitted(token, usagePlan, Configuration.Configuration.get('AWS_ACCESS_KEY', connection), Configuration.Configuration.get('AWS_SECRET_KEY', connection), connection)
    print(permitted)
    name = AmazonWebServices.getName(id, Configuration.Configuration.get('AWS_ACCESS_KEY', connection), Configuration.Configuration.get('AWS_SECRET_KEY', connection))
    print(name)



import pandas as pd
import pandas.api.types as pdt

import pacifico_devel.util.lexikon.src.util.Enumerations as lenm
import pacifico_devel.util.lexikon.src.core.Translator as trl


def translateDataFrame(df: pd.DataFrame,
                       languageInput: lenm.Language,
                       languageOutput: lenm.Language) -> pd.DataFrame:
    """
    Translates all strings in dataframe.

    Args:
        df:
        languageInput:
        languageOutput:

    Returns:

    """
    dfAux = df.copy()
    ##
    if pdt.is_string_dtype(dfAux.index):
        dfAux.index = trl.Translator.translate(text=dfAux.index,
                                               languageInput=languageInput,
                                               languageOutput=languageOutput)
    ##
    if dfAux.index.name is None:
        dfAux.index.name = trl.Translator.translate(text="index",
                                                    languageInput=lenm.Language.English_US,
                                                    languageOutput=languageOutput)
    else:
        dfAux.index.name = trl.Translator.translate(text=dfAux.index.name,
                                                    languageInput=languageInput,
                                                    languageOutput=languageOutput)
    ##
    if pdt.is_string_dtype(dfAux.columns):
        dfAux.columns = trl.Translator.translate(text=dfAux.columns,
                                                 languageInput=languageInput,
                                                 languageOutput=languageOutput)
    for col in dfAux.columns:
        if pdt.is_string_dtype(dfAux[col]):
            dfAux[col] = trl.Translator.translate(text=dfAux[col],
                                                  languageInput=languageInput,
                                                  languageOutput=languageOutput)
    return dfAux



import pathlib as pth

_ROOT_SRC = pth.Path(__file__).parents[1]

PATH_LOGO: pth.Path = pth.Path(__file__).parent / "images" / "Logo_Pacifico.jpg"
PATH_BOKEH_THEME: pth.Path = _ROOT_SRC / "core" / "Style" / "bokeh_themes" / "pacifico_theme.yml"
ROOT_TEMP_IMG_PLOTS: pth.Path = _ROOT_SRC / "core" / "Plot" / "tmp"
PATH_TEMPLATE_EXCEL: pth.Path = _ROOT_SRC / "core" / "Plot" / "tmp" / "output.xlsx"

# if __name__ == '__main__':
#     print(f"{PATH_BOKEH_THEME = }")
#     print(f"{pth.Path(__file__).parents[1] = }")
#
#     print(f"\n {'#' * 20} \n")
#
#     print(f"{PATH_LOGO.is_file() = }")
#     print(f"{PATH_BOKEH_THEME.is_file() = }")
#     print(f"{ROOT_TEMP_IMG_PLOTS.is_dir() = }")



# -*- coding: utf-8 -*-

import abc
import datetime as dt
from abc import ABCMeta, abstractmethod
import pandas as pd
import pathlib as pth
import typing as typ
import re
from docx2pdf import convert
import PyPDF2
import os
import chromedriver_autoinstaller
from openpyxl import load_workbook

from pacifico_devel import Message
from pacifico_devel.api.src.util import Enumerations, Variant

# from pacifico_devel.util.reportTemplate.src.core.Style.Style import Style
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl
# from pacifico_devel.util.reportTemplate.src.core.Plot.Plot import Plot
# import pacifico_devel.util.reportTemplate.src.io.PlotFactory as pltf
import pacifico_devel.util.reportTemplate.src.core.Plot.Plot as bplt
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotAltair.PlotAltairLine as pal
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotAltair.PlotAltairBar as pab
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehLinePercent as pblp
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehLineXticks as pbxt
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehLinePercentXticks as pbpxt
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehLineAndPointed as pblap
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehLine as pbl
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehLineTwoYAxis as pblt
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehPieChart as pbpc
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehBar as pbb
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehMultiBar as pbmb
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokehFigure.PlotBokehStackBar as pbsb
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTable as pt
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableIndex as pti
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableNoIndexes as ptni
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTablePercent as ptpc
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTablePercentIndex as ptpci
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableCarryAndRoll as ptcr
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableColorMap as ptcm
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableOneColorMap as ptocm
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableColorIndex as ptci
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTablePositions as ptp
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableWithDivisions as ptd
import pacifico_devel.util.reportTemplate.src.core.Plot.PlotTable.PlotTableWithDivisionsIndex as ptdi
# import pacifico_devel.util.reportTemplate.src.util.Enumerations as renm
import pacifico_devel.util.reportTemplate.src.util.Report.DateFormatterEn as dtf
import pacifico_devel.util.reportTemplate.src.util.CONFIG as CFG
from pacifico_devel.util.lexikon.src.util.Enumerations import Language
from pacifico_devel.util.lexikon.src.core.Translator import Translator

from pacifico_devel.util.aws import S3
from pacifico_devel.util.cfg import Configuration as cfg

multiDate = typ.Union[dt.date, dt.datetime]


class Report(metaclass=ABCMeta):
    """
    Puts it all together? (format and style, analogue to HTML and CSS.)
    """

    _BASE_PLOT_RESOLUTION: int = 1000
    _EPSILON: float = 2 / 100
    _WRITER = pd.ExcelWriter(CFG.PATH_TEMPLATE_EXCEL, engine='xlsxwriter')

    def __init__(self,
                 pathTemplate: typ.Union[pth.Path, str],
                 languageOutput: typ.Union[Language, str] = Language.Spanish,
                 languageInput: typ.Union[Language, str] = Language.Spanish,
                 date: dt.datetime = dt.datetime.now(),
                 outputFormat: typ.Optional[str] = None):

        chromedriver_autoinstaller.install()

        self._pathTemplate: pth.Path = pth.Path(pathTemplate)
        self._checkPathTemplate()

        if outputFormat is None:
            self._outputFormat: str = pathTemplate.suffix
        else:
            self._outputFormat: str = outputFormat

        if isinstance(languageOutput, str):
            self._languageOutput = Language.fromDeeplString(languageOutput)
        else:
            self._languageOutput = languageOutput

        if isinstance(languageInput, str):
            self._languageInput = Language.fromDeeplString(languageInput)
        else:
            self._languageInput = languageInput

        if not self._pathTemplate.suffix == self._outputFormat:
            raise ValueError(f"File {self._pathTemplate} must be a {self._outputFormat} document.")

        self._date = date
        self._tempPlots: typ.Dict[str, bplt.Plot] = {}

    @property
    def languageOutput(self) -> Language:
        return self._languageOutput

    @property
    def languageInput(self) -> Language:
        return self._languageInput

    @abc.abstractmethod
    def render(self,
               pathOutput: typ.Union[str, pth.Path]) -> None:
        """
        Generates document and saves it locally.

        Args:
            pathOutput:

        Returns:

        """
        pass

    def export(self,
               applicationName: str) -> typ.List[Message.Report]:
        """
        Make a file from abstract reportTemplate and save it to a S3 stated in the filename.

        Args:
           applicationName:

        Returns (str): URL of document.

        """
        self.render(pathOutput=self._tempOutputPath)
        url = self._uploadAndGetUrl(filepathLocal=self._tempOutputPath)
        dateEffective = dt.datetime.now()
        message = self._getMessage(url=url,
                                   applicationName=applicationName,
                                   dateEffective=dateEffective)
        return [message]

    def export2excel(self, applicationName: str, clean: bool = True):
        self._WRITER.close()
        pathExcel = CFG.PATH_TEMPLATE_EXCEL
        url = self._uploadAndGetUrl(filepathLocal=pathExcel)
        dateEffective = dt.datetime.now()
        message = self._getMessage(url=url,
                                   applicationName=applicationName,
                                   dateEffective=dateEffective)
        return [message]

    def exportPDF(self, applicationName: str, clean: bool = True):
        self.render(pathOutput=self._tempOutputPath)
        convert(self._tempOutputPath)
        pathPDF = pth.Path(self._tempOutputPath.parent) / (self._tempOutputPath.stem + ".pdf")
        if clean:
            self.cleanPDF(pathPDF)
        url = self._uploadAndGetUrl(filepathLocal=pathPDF)
        dateEffective = dt.datetime.now()
        message = self._getMessage(url=url,
                                   applicationName=applicationName,
                                   dateEffective=dateEffective)
        return [message]

    def cleanPDF(self, filePathReader: typ.Union[str, pth.Path]):
        fileName: str = self._pathTemplate.stem + "_temp1.pdf"
        filePathWriter = self._pathTemplate.parent / fileName
        writer = PyPDF2.PdfFileWriter()
        file = open(filePathReader, 'rb')
        reader = PyPDF2.PdfFileReader(file)
        for index, page in enumerate(reader.pages):
            texts = [text for text in page.extractText().split('\n') if not re.search("^\s*$", text)]
            if len(texts) > 2 or index == 0:
                writer.addPage(page)
        outputFile = open(filePathWriter, "wb")
        writer.write(outputFile)
        file.close()
        os.remove(filePathReader)
        outputFile.close()
        os.rename(filePathWriter, filePathReader)

    def _deleteAllTempPlots(self) -> None:
        """
        deletes all temporary plot images

        Returns:

        """
        [plot.deleteImage()
         for plot in self._tempPlots.values()]

    def _getMessage(self,
                    url: str,
                    applicationName: str,
                    dateEffective: dt.datetime) -> Message.Report:
        """
        Returns information in API format (Message.Report).

        Args:
            url:
            applicationName:
            dateEffective:

        Returns:

        """
        # document, chapter, section, subsection, paragraph = self.splitApplicationName(applicationName=applicationName)
        keys: typ.List[str] = ["document", "chapter", "section", "subsection", "paragraph"]
        values: typ.List[str] = self._splitApplicationName(applicationName=applicationName)

        if not 1 <= len(values) <= 4:
            msg1 = "Expecting name to be split into list of between 1 and 4 elements, "
            msg2 = f"but returned: {values}"
            raise ValueError(msg1 + msg2)

        values.append(self._outputFormat)  # put format last
        argDict: typ.Dict[str, typ.Union[str, dt.datetime, Variant.Variant]] = {k: v
                                                                                for k, v in zip(keys, values)}
        argDict["item"] = "report"
        argDict["dateEffective"] = dateEffective
        variant: Variant.Variant = Variant.Variant(value=url,
                                                   valueType=Enumerations.ValueType.ValueType_Url)
        print(f"URL: {url}.")
        argDict["variant"] = variant

        message: Message.Report = Message.Report(**argDict)
        return message

    @abstractmethod
    def _setString(self,
                   tag,
                   value: str) -> None:
        """
        Adds string to document.

        Args:
            tag:
            value:

        Returns:

        """
        pass

    @abstractmethod
    def setImage(self,
                 tag: str,
                 path: typ.Union[str, pth.Path],
                 width: typ.Optional[float] = None,
                 height: typ.Optional[float] = None) -> None:
        """
        Sets an image.

        Returns:

        """

        pass

    def setString(self,
                  tag: str,
                  value: str,
                  translate: bool = False) -> None:  # could be enumeration or beautifulSoup:
        """
        For bond mnemo (for example).

        Args:
            value:
            tag:
            translate:

        Returns:

        """
        valueString = value
        if translate:
            print("\n" + "#" * 20)
            print("TRANSLATING...")
            msg1 = f"TEXT TO TRANSLATE FROM {self._languageInput.deeplStringInput} to {self._languageOutput.deeplStringOutput}:"
            msg2 = f"\n\t{valueString}"
            print(msg1 + msg2)
            valueString = Translator.translate(text=valueString,
                                               languageInput=self.languageInput,
                                               languageOutput=self.languageOutput)
            print(f"TRANSLATION : \n\t{valueString}")
            print("TRANSLATING... DONE!")
            print("#" * 20 + "\n")
        self._setString(value=valueString,
                        tag=tag)

    def setText(self,
                tag: str,
                value: str):
        """
        For Commentary (for example)

        Args:
            value:
            tag:

        Returns:

        """
        return self.setString(value=value,
                              tag=tag,
                              translate=True)

    def setInteger(self,
                   tag: str,
                   value: int) -> None:
        self.setString(value=stl.Style.convertIntToString(value,
                                                          self.languageOutput),
                       tag=tag)

    def setFloat(self,
                 tag: str,
                 value: float,
                 numberDecimals: int,
                 isPercentage: bool = False) -> None:
        """
        Same as above.

        Args:
            value:
            numberDecimals:
            tag:
            isPercentage:

        Returns:

        """
        valueString = stl.Style.convertFloatToString(value=value,
                                                     numberDecimals=numberDecimals,
                                                     language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setLongDate(self,
                    tag: str,
                    value: typ.Optional[multiDate] = None,
                    ) -> None:

        if value is None:
            longDateStringEn = dtf.DateFormatterEn.getLongDateStringEn(self._date)
        else:
            longDateStringEn = dtf.DateFormatterEn.getLongDateStringEn(value)

        longDateString = Translator.translate(text=longDateStringEn,
                                              languageInput=Language.English_US,
                                              languageOutput=self.languageOutput)
        self.setString(value=longDateString,
                       tag=tag,
                       translate=False)

    def setDateTime(self,
                    value: dt.datetime,
                    tag: str) -> None:
        valueString = stl.Style.convertDateTimeToString(value=value,
                                                        language=self.languageOutput)
        self.setString(value=valueString,
                       tag=tag)

    def setDate(self,
                value: dt.date,
                tag: str) -> None:
        """


        Args:
            value:
            tag:

        Returns:

        """
        valueString = stl.Style.convertDateToString(value,
                                                    language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setDateMonth(self,
                     value: dt.date,
                     tag: str) -> None:
        """
        YYYY/MM

        Args:
            value:
            tag:

        Returns:

        """
        valueString = stl.Style.convertDateMonthToString(value,
                                                         language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setDateQuarter(self,
                       value: dt.date,
                       tag: str) -> None:
        """
        YYYY

        Args:
            value:
            tag:

        Returns:

        """
        valueString = stl.Style.convertDateQuarterToString(value,
                                                           language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setDateYear(self,
                    value: dt.date,
                    tag: str) -> None:
        """
        YYYY

        Args:
            value:
            tag:

        Returns:

        """
        valueString = stl.Style.convertDateYearToString(value)

        self.setString(value=valueString,
                       tag=tag)

    # def setPlot(self,
    #             tag: str,
    #             df: pd.DataFrame,
    #             plotKind: typ.Union[renm.PlotKind, str],
    #             title: str,
    #             shape: typ.Tuple[float, float],
    #             backend: typ.Union[renm.Backend, str, None] = None):  # width, height
    #
    #     if backend is None:
    #         backend = self._plotBackend
    #
    #     if isinstance(plotKind, str):
    #         plotKind = renm.PlotKind[plotKind.upper()]
    #
    #     p = pltf.PlotFactory.getPlot(plotKind=plotKind,
    #                                  languageInput=self._languageInput,
    #                                  languageOutput=self._languageOutput)
    #
    #     if title:
    #         titleTranslated = Translator.translate(title,
    #                                                languageInput=self._languageInput,
    #                                                languageOutput=self._languageOutput)
    #     else:
    #         titleTranslated = title
    #
    #     p.makePlot(df=df,
    #                title=titleTranslated,
    #                shape=tuple(int(dim * self._BASE_PLOT_RESOLUTION)
    #                            for dim in shape
    #                            ),
    #                backend=backend)
    #     self.setImage(tag=tag,
    #                   path=p.pathImage,
    #                   width=shape[0],
    #                   height=shape[1])
    #     self._tempPlots[tag] = p

    # SET PLOTS

    def _set_plot(self,
                  plot: bplt.Plot,
                  tag: str,
                  shape: typ.Tuple[float, float]) -> None:
        """
        Creates image from plot object, then sets the image into report and finally inserts plot into
        an internal dictionary for later access and erasure.

        Args:
            plot:
            tag:
            shape:

        Returns:

        """
        plot.saveImage()
        plot.df_to_excel(self._WRITER, tag)

        self.setImage(tag=tag,
                      path=plot.pathImage,
                      width=shape[0],
                      height=shape[1])
        self._tempPlots[tag] = plot

    def setPlotLine(self,
                    tag: str,
                    df: pd.DataFrame,
                    title: str,
                    shape: typ.Tuple[float, float]) -> None:

        p = pal.PlotAltairLine(df=df,
                               shape=shape,
                               languageInput=self.languageInput,
                               languageOutput=self.languageOutput,
                               basePlotResolution=self._BASE_PLOT_RESOLUTION,
                               title=title)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotLinePercent(self,
                           tag: str,
                           df: pd.DataFrame,
                           title: str,
                           shape: typ.Tuple[float, float]) -> None:

        p = pblp.PlotBokehLinePercent(df=df,
                                      shape=shape,
                                      languageInput=self.languageInput,
                                      languageOutput=self.languageOutput,
                                      basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                      title=title)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotLineXticks(self,
                          tag: str,
                          df: pd.DataFrame,
                          title: str,
                          shape: typ.Tuple[float, float],
                          xTicks: list = None) -> None:

        p = pbxt.PlotBokehLineXticks(df=df,
                                     shape=shape,
                                     languageInput=self.languageInput,
                                     languageOutput=self.languageOutput,
                                     basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                     title=title,
                                     xTicks=xTicks)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotLinePercentXticks(self,
                                 tag: str,
                                 df: pd.DataFrame,
                                 title: str,
                                 shape: typ.Tuple[float, float],
                                 xTicks: list = None) -> None:

        p = pbpxt.PlotBokehLinePercentXticks(df=df,
                                             shape=shape,
                                             languageInput=self.languageInput,
                                             languageOutput=self.languageOutput,
                                             basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                             title=title,
                                             xTicks=xTicks)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotBokehLine(self,
                         tag: str,
                         df: pd.DataFrame,
                         title: str,
                         shape: typ.Tuple[float, float]) -> None:

        p = pbl.PlotBokehLine(df=df,
                              shape=shape,
                              languageInput=self.languageInput,
                              languageOutput=self.languageOutput,
                              basePlotResolution=self._BASE_PLOT_RESOLUTION,
                              title=title)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotBokehLineAndPointed(self,
                                   tag: str,
                                   df: pd.DataFrame,
                                   title: str,
                                   shape: typ.Tuple[float, float]) -> None:

        p = pblap.PlotBokehLineAndPointed(df=df,
                                          shape=shape,
                                          languageInput=self.languageInput,
                                          languageOutput=self.languageOutput,
                                          basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                          title=title)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotLineTwoAxis(self,
                           tag: str,
                           df: pd.DataFrame,
                           title: str,
                           shape: typ.Tuple[float, float]) -> None:

        p = pblt.PlotBokehLineTwoYAxis(df=df,
                                       shape=shape,
                                       languageInput=self.languageInput,
                                       languageOutput=self.languageOutput,
                                       basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                       title=title)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotPieChart(self,
                        tag: str,
                        df: pd.DataFrame,
                        title: str,
                        shape: typ.Tuple[float, float]) -> None:

        p = pbpc.PlotBokehPieChart(df=df,
                                   shape=shape,
                                   languageInput=self.languageInput,
                                   languageOutput=self.languageOutput,
                                   basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                   title=title)

        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotBar(self,
                   tag: str,
                   df: pd.DataFrame,
                   title: str,
                   shape: typ.Tuple[float, float]) -> None:

        p = pab.PlotAltairBar(df=df,
                              shape=shape,
                              languageInput=self.languageInput,
                              languageOutput=self.languageOutput,
                              basePlotResolution=self._BASE_PLOT_RESOLUTION,
                              title=title)

        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotStackBar(self,
                        tag: str,
                        df: pd.DataFrame,
                        title: str,
                        shape: typ.Tuple[float, float]) -> None:

        p = pbsb.PlotBokehStackBar(df=df,
                                   shape=shape,
                                   languageInput=self.languageInput,
                                   languageOutput=self.languageOutput,
                                   basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                   title=title)

        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotBar2(self,
                    tag: str,
                    df: pd.DataFrame,
                    title: str,
                    shape: typ.Tuple[float, float]) -> None:

        p = pbb.PlotBokehBar(df=df,
                             shape=shape,
                             languageInput=self.languageInput,
                             languageOutput=self.languageOutput,
                             basePlotResolution=self._BASE_PLOT_RESOLUTION,
                             title=title)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setPlotMultiBar2(self,
                         tag: str,
                         df: pd.DataFrame,
                         title: str,
                         shape: typ.Tuple[float, float]) -> None:
        p = pbmb.PlotBokehMultiBar(df=df,
                                   shape=shape,
                                   languageInput=self.languageInput,
                                   languageOutput=self.languageOutput,
                                   basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                   title=title)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTable(self,
                 tag: str,
                 df: pd.DataFrame,
                 shape: typ.Tuple[float, float],
                 formatter=None) -> None:

        p = pt.PlotTable(df=df,
                         shape=shape,
                         languageInput=self.languageInput,
                         languageOutput=self.languageOutput,
                         basePlotResolution=self._BASE_PLOT_RESOLUTION,
                         formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableNoIndexes(self,
                          tag: str,
                          df: pd.DataFrame,
                          shape: typ.Tuple[float, float],
                          formatter=None) -> None:

        p = ptni.PlotTableNoIndexes(df=df,
                                    shape=shape,
                                    languageInput=self.languageInput,
                                    languageOutput=self.languageOutput,
                                    basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                    formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTablePercent(self,
                        tag: str,
                        df: pd.DataFrame,
                        shape: typ.Tuple[float, float]) -> None:

        p = ptpc.PlotTablePercent(df=df,
                                  shape=shape,
                                  languageInput=self.languageInput,
                                  languageOutput=self.languageOutput,
                                  basePlotResolution=self._BASE_PLOT_RESOLUTION)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTablePercentIndex(self,
                             tag: str,
                             df: pd.DataFrame,
                             shape: typ.Tuple[float, float]) -> None:

        p = ptpci.PlotTablePercentIndex(df=df,
                                        shape=shape,
                                        languageInput=self.languageInput,
                                        languageOutput=self.languageOutput,
                                        basePlotResolution=self._BASE_PLOT_RESOLUTION)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableCarryAndRoll(self,
                             tag: str,
                             df: pd.DataFrame,
                             shape: typ.Tuple[float, float],
                             formatter=None) -> None:

        p = ptcr.PlotTableCarryAndRoll(df=df,
                                       shape=shape,
                                       languageInput=self.languageInput,
                                       languageOutput=self.languageOutput,
                                       basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                       formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableColorIndex(self,
                           tag: str,
                           df: pd.DataFrame,
                           shape: typ.Tuple[float, float],
                           formatter=None) -> None:

        p = ptci.PlotTableColorIndex(df=df,
                                     shape=shape,
                                     languageInput=self.languageInput,
                                     languageOutput=self.languageOutput,
                                     basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                     formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableIndex(self,
                      tag: str,
                      df: pd.DataFrame,
                      shape: typ.Tuple[float, float],
                      formatter=None) -> None:

        p = pti.PlotTableIndex(df=df,
                               shape=shape,
                               languageInput=self.languageInput,
                               languageOutput=self.languageOutput,
                               basePlotResolution=self._BASE_PLOT_RESOLUTION,
                               formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableColorMap(self,
                         tag: str,
                         df: pd.DataFrame,
                         shape: typ.Tuple[float, float],
                         formatter=None) -> None:

        p = ptcm.PlotTableColorMap(df=df,
                                   shape=shape,
                                   languageInput=self.languageInput,
                                   languageOutput=self.languageOutput,
                                   basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                   formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableOneColorMap(self,
                            tag: str,
                            df: pd.DataFrame,
                            shape: typ.Tuple[float, float],
                            formatter=None) -> None:

        p = ptocm.PlotTableOneColorMap(df=df,
                                       shape=shape,
                                       languageInput=self.languageInput,
                                       languageOutput=self.languageOutput,
                                       basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                       formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTablePositions(self,
                          tag: str,
                          df: pd.DataFrame,
                          shape: typ.Tuple[float, float],
                          formatter=None) -> None:

        p = ptp.PlotTablePositions(df=df,
                                   shape=shape,
                                   languageInput=self.languageInput,
                                   languageOutput=self.languageOutput,
                                   basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                   formatter=formatter)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableWithDivisions(self,
                              tag: str,
                              df: pd.DataFrame,
                              shape: typ.Tuple[float, float],
                              formatter=None,
                              columnsSubset: list = []) -> None:

        p = ptd.PlotTableWithDivisions(df=df,
                                       shape=shape,
                                       languageInput=self.languageInput,
                                       languageOutput=self.languageOutput,
                                       basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                       formatter=formatter,
                                       columnsSubset=columnsSubset)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    def setTableWithDivisionsIndex(self,
                                   tag: str,
                                   df: pd.DataFrame,
                                   shape: typ.Tuple[float, float],
                                   formatter=None,
                                   columnsSubset: list = []) -> None:

        p = ptdi.PlotTableWithDivisionsIndex(df=df,
                                             shape=shape,
                                             languageInput=self.languageInput,
                                             languageOutput=self.languageOutput,
                                             basePlotResolution=self._BASE_PLOT_RESOLUTION,
                                             formatter=formatter,
                                             columnsSubset=columnsSubset)
        self._set_plot(plot=p,
                       tag=tag,
                       shape=shape)

    #
    # def setTable(self,
    #              df: pd.DataFrame,
    #              tag: str,
    #              shape: typ.Tuple[int, int] = (10, 10),
    #              formatter: dict = None):
    #     """
    #     PLAN: df (Done through Style) -> HTML -> IMAGE -> Paste in reportTemplate.
    #
    #     Args:
    #         df:
    #         tag:
    #         shape:
    #
    #     Returns:
    #
    #     """
    #     p = pltf.PlotFactory.getPlot(plotKind=renm.PlotKind.TABLE_PLOT,
    #                                  languageInput=self._languageInput,
    #                                  languageOutput=self._languageOutput)
    #
    #     p.makePlot(df=df,
    #                shape=tuple(int(dim * self._BASE_PLOT_RESOLUTION)
    #                            for dim in shape),
    #                backend=self._plotBackend,
    #                formatter=formatter)
    #     self.setImage(tag=tag,
    #                   path=p.pathImage,
    #                   width=shape[0],
    #                   height=shape[1])
    #     self._tempPlots[tag] = p

    @property
    def _tempOutputPath(self) -> pth.Path:
        fileName: str = self._pathTemplate.stem + "_temp" + self._outputFormat
        return self._pathTemplate.parent / fileName

    def _checkPathTemplate(self) -> None:
        """
        Assures the input template's path exists.

        """
        if not self._pathTemplate.is_file():
            msg = f"No such template path: {self._pathTemplate}. Please initialize with a valid template path."
            raise FileNotFoundError(msg)

    def _checkOuputFormat(self) -> None:
        """
        Assures input template's format is the output's file format.

        """
        if not self._pathTemplate.suffix == self._outputFormat:
            raise ValueError(f"File {self._pathTemplate} must be a {self._outputFormat} document.")

    def _uploadAndGetUrl(self, filepathLocal: typ.Union[str, pth.Path]) -> str:

        """
        Uploads file to S3 and returns URL.

        Args:
            filepathLocal:

        Returns:

        """

        if isinstance(filepathLocal, str):
            filepathLocal = pth.Path(filepathLocal)
        filepathBucket: pth.Path = self._getFilePathUpload(filePathLocal=filepathLocal)
        bucketName = "media-pacifico"
        awsAccessKey = cfg.get("AWS_ACCESS_KEY_PACIFICO")
        awsSecretKey = cfg.get("AWS_SECRET_KEY_PACIFICO")
        S3.writeFileToBucket(filepathLocal.as_posix(),
                             bucketName,
                             awsAccessKey,
                             awsSecretKey,
                             filepathBucket.as_posix())  # also erases file
        url = S3.getUrl(filepathBucket.as_posix(),
                        bucketName,
                        awsAccessKey=awsAccessKey,
                        awsSecretKey=awsSecretKey)
        return url

    @staticmethod
    def _splitApplicationName(applicationName: str) -> typ.List[str]:
        splitName: typ.List[str] = re.findall("[A-Z][a-z]+",
                                              applicationName)  # NEEDS TO INCLUDE CAPITAL LETTERS FOLLOWING
        splitName = [s for s in splitName if s != "Application"]
        if len(splitName) == 0:
            splitName = ["Report"]
        if len(splitName) >= 5:
            sFinal = " ".join(splitName[4:])
            splitName = splitName[:4]
            splitName[4] = sFinal
        return splitName

    @staticmethod
    def _getFilePathUpload(filePathLocal: pth.Path) -> pth.Path:
        stringTime: str = dt.datetime.now().strftime("%Y.%m.%d_%H.%M.%S.%f")
        fileNameUpload = filePathLocal.stem + "_" + stringTime + filePathLocal.suffix
        return pth.Path(fileNameUpload)


if __name__ == '__main__':
    # print(__file__)
    # print(pth.Path(pth.Path(__file__)))
    # test = "ApplicationStuffABTesting"
    # print(Report.splitApplicationName(test))
    filepathBucket = "template_temp_2022.04.22_17.17.36.906014.docx"
    filepathLocal = "template_temp_2022.04.22_17.17.36.906014.docx"
    bucketName = "media-pacifico"
    awsAccessKey = cfg.get("AWS_ACCESS_KEY_PACIFICO")
    awsSecretKey = cfg.get("AWS_SECRET_KEY_PACIFICO")
    # S3.writeFileToBucket(filepathLocal.as_posix(),
    #                      bucketName,
    #                      awsAccessKey,
    #                      awsSecretKey,
    #                      filepathBucket.as_posix())  # also erases file
    url = S3.getUrl(filepathBucket,
                    bucketName,
                    awsAccessKey=awsAccessKey,
                    awsSecretKey=awsSecretKey)
    print(url)



from pacifico_devel.util.reportTemplate.src.core.Plot.PlotBokeh.PlotBokeh import PlotBokeh
import pacifico_devel.util.reportTemplate.src.util.PacificoPlotColors as pc
import pacifico_devel.util.lexikon.src.util.Enumerations as lenm
import pacifico_devel.util.reportTemplate.src.core.Style.Style as stl

import bokeh.plotting as bpl
from bokeh.models import ColumnDataSource
from bokeh.plotting import figure
from bokeh.transform import dodge
from bokeh.io import curdoc
from bokeh.themes import Theme
from bokeh.models import NumeralTickFormatter
import pandas as pd
import typing as typ
from bokeh.plotting import output_file, show
from bokeh.io import export_png


class PlotBokehMultiBar(PlotBokeh):
    def __init__(self,
                 df: pd.DataFrame,
                 shape: typ.Tuple[float, float],
                 languageInput: lenm.Language,
                 languageOutput: lenm.Language,
                 basePlotResolution: int,
                 style: typ.Optional[stl.Style] = None,
                 title: str = None):
        super(PlotBokehMultiBar, self).__init__(df=df, shape=shape, languageInput=languageInput,
                                                languageOutput=languageOutput, basePlotResolution=basePlotResolution,
                                                style=style)
        self._title = title

    def _makeChart(self) -> bpl.Figure:
        maxValue = self.df.iloc[:, 1:].max().max()
        data = self.df.to_dict(orient="list")
        positions, width = PlotBokehMultiBar.plotDistribution(len(list(data.keys())) - 1)
        source = ColumnDataSource(data=data)
        plot = figure(x_range=data['xAxisCategory'],
                      y_range=(0, maxValue * 1.1),
                      width=self._shape[0],
                      height=self._shape[1],
                      toolbar_location=None)
        plot.title = self._title
        categories = list(data.keys())
        categories.remove('xAxisCategory')
        for category, color, position in zip(categories, pc.PacificoPlotColors, positions):
            plot.vbar(x=dodge('xAxisCategory', position, range=plot.x_range), top=category, width=width,
                      source=source, color=color.getCodeString(), legend_label=category)
        if len(data['xAxisCategory']) > 12:
            plot.xaxis.major_label_orientation = "vertical"
        self.style(plot)
        self.legendStyle(plot, 'horizontal', 'below')
        # show(plot)
        # export_png(plot, filename="bokeh_plot.png")
        return plot

    def legendStyle(self, plot, orientation: str = 'vertical', layout: str = 'right'):
        new_legend = plot.legend[0]
        plot.add_layout(new_legend, layout)  # 'below'
        plot.legend.orientation = orientation  # 'horizontal'
        plot.legend.location = "center"
        plot.legend.border_line_alpha = 0
        plot.legend.background_fill_color = None
        plot.xgrid.grid_line_color = None
        plot.legend.click_policy = "hide"
        plot.legend.label_text_font = "klartext mono"
        plot.legend.label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        return plot

    def style(self, plot):
        # TITLE
        plot.title.text_color = pc.PacificoPlotColors.BLUE.getCodeString()
        plot.title.text_font_size = "20px"
        plot.title.text_font = "klartext mono"
        # plot.title.text_font_style = "italic"
        # LABELS
        plot.axis.axis_label_text_font = "klartext mono"
        plot.axis.axis_label_text_font_size = "15px"
        plot.axis.axis_label_text_font_style = 'normal'
        plot.axis.axis_label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        plot.axis.major_label_text_color = pc.PacificoPlotColors.DARK_GREY.getCodeString()
        # AXIS
        plot.xaxis.axis_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.axis.major_tick_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.xaxis.major_tick_line_width = 1
        plot.axis.minor_tick_line_color = pc.PacificoPlotColors.LIGHT_GREY.getCodeString()
        plot.yaxis.minor_tick_line_color = None
        plot.yaxis.axis_line_color = None
        # TICKS
        # plot.xaxis.major_label_orientation = np.pi / 2
        # plot.xaxis.major_label_orientation = "vertical"
        # BACKGROUND
        plot.xgrid.grid_line_color = None
        plot.outline_line_color = None
        plot.yaxis[0].formatter = NumeralTickFormatter(format="y{0%} ")
        return plot

    @staticmethod
    def debugMakeChart(data, shape, title='') -> bpl.Figure:
        # output_file("PlotBokehMultiBar.html")
        maxValue = data.iloc[:, 1:].max().max()
        data = data.to_dict(orient="list")
        positions, width = PlotBokehMultiBar.plotDistribution(len(list(data.keys())) - 1)
        source = ColumnDataSource(data=data)
        plot = figure(x_range=data['xAxisCategory'],
                      y_range=(0, maxValue * 1.1),
                      width=shape[0],
                      height=shape[1],
                      toolbar_location=None)
        plot.title = title
        categories = list(data.keys())
        categories.remove('xAxisCategory')
        for category, color, position in zip(categories, pc.PacificoPlotColors, positions):
            plot.vbar(x=dodge('xAxisCategory', position, range=plot.x_range), top=category, width=width,
                      source=source, color=color.getCodeString(), legend_label=category)
        if len(data['xAxisCategory']) >= 12:
            plot.xaxis.major_label_orientation = "vertical"
        show(plot)
        export_png(plot, filename="bokeh_plot.png")
        return plot

    @staticmethod
    def plotDistribution(numberBarCategory):
        if numberBarCategory == 6:
            positions = [-0.45, -0.3, -0.15, 0, 0.15, 0.3, 0.45]
            width = 0.07
            return positions, width
        if numberBarCategory == 5:
            positions = [-0.3, -0.15, 0, 0.15, 0.3]
            width = 0.07
            return positions, width
        if numberBarCategory == 4:
            positions = [-0.15, -0.5, 0.5, 0.15]
            width = 0.08
            return positions, width
        if numberBarCategory == 3:
            positions = [-0.1, 0, 0.1]
            width = 0.085
            return positions, width
        if numberBarCategory == 2:
            positions = [-0.1, 0.1]
            width = 0.1
            return positions, width


if __name__ == '__main__':
    from datetime import datetime
    from Applications.ApplicationReportChileUnemployment.src.core.Data import Data
    from pacifico_devel.util.lexikon.src.util.Enumerations import Language
    import chromedriver_autoinstaller

    chromedriver_autoinstaller.install()
    data = Data(datetime.now())
    data.load('KmK9FqgJ5h1pSNjh1vhha6ul30vXBOBiMVlmplXf', ['60703000-6', 'F049.CES.TAS.UCH3.65.T'])
    data.computeUnemployment()
    data = data.getData()
    months = {'ene': 1, 'feb': 2, 'mar': 3, 'abr': 4, 'may': 5, 'jun': 6, 'jul': 7,
              'ago': 8, 'sep': 9, 'oct': 10, 'nov': 11, 'dic': 12}
    dataDict = {}
    years = list(set(data['Año']))
    years.sort()
    for year in years[-5:]:
        dataDict[str(year)] = data[data['Año'] == year]['Desempleo (Total)'].tolist()
    dataPlot = {**{'xAxisCategory': list(months.keys())}, **dataDict}

    dataPlot = pd.DataFrame(dataPlot)

    PlotBokehMultiBar(df=dataPlot,
                      shape=(900, 500),
                      languageInput=Language.fromDeeplString("ES"),
                      languageOutput=Language.fromDeeplString("ES"),
                      basePlotResolution=1000,
                      style=None,
                      title='')._makeChart()  # debugMakeChart(dataPlot, (900, 500))



from enum import Enum, auto
import pathlib as pth


class PlotKind(Enum):
    LINE_PLOT = auto()
    BAR_PLOT = auto()
    TABLE_PLOT = auto()
    MULTI_BAR_PLOT = auto()


class Backend(Enum):
    BOKEH = auto()
    ALTAIR = auto()


class BokehTheme(Enum):
    THEME_1 = auto()
    THEME_2 = auto()

    @property
    def path(self) -> pth.Path:
        if self is self.__class__.THEME_1:
            return pth.Path(__file__).parents[1] / "core" / "Style" / "bokeh_themes" / "pacifico_theme.yml"
        elif self is self.__class__.THEME_2:
            # INSERT PATH HERE
            pass


if __name__ == '__main__':
    bt = BokehTheme.THEME_1
    print(bt.path)
    print(bt.path.is_file())
